"""The AWS Lambda emmaa-model-update definition.

This file contains the function that starts model update cycle. It must be
placed on AWS Lambda, which can either be done manually (not recommended) or by
running:

$ python update_lambda.py model_updates.py emmaa-model-update

in this directory.
"""

import boto3
from datetime import datetime

JOB_DEF = 'emmaa_jobdef'
QUEUE = 'emmaa-models-update-test'
PROJECT = 'aske'
PURPOSE = 'update-emmaa-models'
BRANCH = 'origin/master'


def lambda_handler(event, context):
    """Create a batch job to update models on s3 and NDEx.

    This function is designed to be placed on AWS Lambda, taking the event and
    context arguments that are passed. Note that this function must always have
    the same parameters, even if any or all of them are unused, because we do
    not have control over what Lambda sends as parameters. Event parameter is
    used to pass model_name argument.

    This Lambda function is configured to be invoked by emmaa-update-pipeline
    Lambda function.

    See the top of the page for the Lambda update procedure.

    Parameters
    ----------
    event : dict
        A dictionary containing metadata regarding the triggering event.
    context : object
        This is an object containing potentially useful context provided by
        Lambda. See the documentation cited above for details.

    Returns
    -------
    ret : dict
        A dict containing 'statusCode', with a valid HTTP status code, 'result',
        and 'job_id' to be returned to Lambda.
    """
    model_name = event['model']
    batch = boto3.client('batch')
    core_command = 'bash scripts/git_and_run.sh'
    if BRANCH is not None:
        core_command += f' --branch {BRANCH} '
    core_command += (f'python scripts/run_model_update.py --model {model_name}')
    print(core_command)
    cont_overrides = {
        'command': ['python', '-m', 'indra.util.aws', 'run_in_batch',
                    '--project', PROJECT, '--purpose', PURPOSE,
                    core_command]
        }
    now_str = datetime.utcnow().strftime('%Y%m%d_%H%M%S')
    ret = batch.submit_job(jobName=f'{model_name}_update_{now_str}',
                           jobQueue=QUEUE, jobDefinition=JOB_DEF,
                           containerOverrides=cont_overrides)
    job_id = ret['jobId']

    return {'statusCode': 200, 'result': 'SUCCESS', 'job_id': job_id}



from emmaa.priors import SearchTerm
from emmaa.model import save_config_to_s3


def make_search_terms(terms, ontology_file):
    """Make SearchTerm objects standardized to a given ontology from terms.

    Parameters
    ----------
    terms : list[str]
        A list of terms corresponding to suffixes of entries in the ontology.
    ontology_file : str
        A path to a file containing ontology.

    Returns
    -------
    search_terms : set
        A set of SearchTerm objects constructed from given terms and ontology
        having standardized names.
    """
    search_terms = set()
    with open(ontology_file, 'r') as f:
        lines = f.readlines()
    ontologies = []
    for line in lines:
        links = line.split('> <')
        link = links[0]
        ont_start = link.find('UN')
        ont = link[ont_start:]
        ontologies.append(ont)
    for ont in ontologies:
        for term in terms:
            if ont.endswith(term):
                search_term = term.replace('_', ' ')
                name = search_term.capitalize()
                st = SearchTerm(type='concept', name=name, db_refs={'UN': ont},
                                search_term='\"%s\"' % search_term)
                search_terms.add(st)
    return search_terms


def make_config(search_terms, human_readable_name, description,
                short_name, ndex_network=None, save_to_s3=False):
    """Make a config file for WorldModelers models and optionally save to S3."""
    config = {}
    config['ndex'] = {'network': ndex_network if ndex_network else ''}
    config['human_readable_name'] = human_readable_name
    config['search_terms'] = [st.to_json() for st in search_terms]
    config['test'] = {'statement_checking': {
                      'max_path_length': 5,
                      'max_paths': 1},
                      'test_corpus': 'world_modelers_tests.pkl'}
    config['assembly'] = {'skip_map_grounding': True,
                          'skip_filter_human': True,
                          'skip_map_sequence': True,
                          'belief_cutoff': 0.8,
                          'filter_ungrounded': True,
                          'score_threshold': 0.7,
                          'filter_relevance': 'prior_one',
                          'standardize_names': True,
                          'preassembly_mode': 'wm'}
    config['reading'] = {'literature_source': 'elsevier',
                         'reader': 'elsevier_eidos'}
    config['description'] = description
    if save_to_s3:
        save_config_to_s3(short_name, config)
    return config



"""This module contains classes to generate prior networks."""


class SearchTerm(object):
    """Represents a search term to be used in a model configuration.

    Parameters
    ----------
    type : str
        The type of search term, e.g. gene, bioprocess, other
    name : str
        The name of the search term, is equivalent to an Agent name
    db_refs : dict
        A dict of database references for the given term, is similar
        to an Agent db_refs dict
    search_term : str
        The actual search term to us for searching PubMed
    """
    def __init__(self, type, name, db_refs, search_term):
        self.type = type
        self.name = name
        self.db_refs = db_refs
        self.search_term = search_term

    def __str__(self):
        return f'SearchTerm({self.type}, {self.name})'

    def __repr__(self):
        return str(self)

    def __hash__(self):
        return hash((self.type, self.name, tuple(sorted(self.db_refs.items())),
                     self.search_term))

    def to_json(self):
        """Return search term as JSON."""
        jd = {'type': self.type,
              'name': self.name,
              'db_refs': self.db_refs,
              'search_term': self.search_term}
        return jd

    @classmethod
    def from_json(cls, jd):
        """Return a SearchTerm object from JSON."""
        return SearchTerm(**jd)

    def __eq__(self, other):
        return self.type == other.type and self.name == other.name and \
            self.db_refs == other.db_refs and \
            self.search_term == other.search_term


def get_drugs_for_gene(stmts, hgnc_id):
    """Get list of drugs that target a gene

    Parameters
    ----------
    stmts : list of :py:class:`indra.statements.Statement`
        List of INDRA statements with a drug as subject
    hgnc_id : str
        HGNC id for a gene

    Returns
    -------
    drugs_for_gene : list of :py:class:`emmaa.priors.SearchTerm`
        List of search terms for drugs targeting the input gene
    """
    drugs_for_gene = []
    for stmt in stmts:
        if stmt.obj.db_refs.get('HGNC') == hgnc_id:
            term = SearchTerm(type='drug', name=stmt.subj.name,
                              db_refs=stmt.subj.db_refs,
                              search_term=f'"{stmt.subj.name}"')
            drugs_for_gene.append(term)
    return drugs_for_gene



class EmmaaStatement(object):
    """Represents an EMMAA Statement.

    Parameters
    ----------
    stmt : indra.statements.Statement
        An INDRA Statement
    date : datetime
        A datetime object that is attached to the Statement. Typically
        represents the time at which the Statement was created.
    search_terms : list[emmaa.priors.SearchTerm]
        The slist of search terms that led to the creation of the Statement.
    """
    def __init__(self, stmt, date, search_terms):
        self.stmt = stmt
        self.date = date
        self.search_terms = search_terms

    def __repr__(self):
        return '%s(%s, %s, %s)' % (self.__class__.__name__, self.stmt,
                                   self.date, self.search_terms)

    def to_json(self):
        output_json = {'search_terms': [st.to_json() for st
                                        in self.search_terms],
                       'date': self.date.strftime('%Y-%m-%d-%H-%M-%S')}
        # Get json representation of statement
        json_stmt = self.stmt.to_json(use_sbo=False)
        # Stringify source hashes: JavaScript can't handle int's of length > 16
        for ev in json_stmt['evidence']:
            ev['source_hash'] = str(ev['source_hash'])
        output_json['stmt'] = json_stmt
        return output_json


def to_emmaa_stmts(stmt_list, date, search_terms):
    """Make EMMAA statements from INDRA Statements with the given metadata."""
    emmaa_stmts = []
    for indra_stmt in stmt_list:
        es = EmmaaStatement(indra_stmt, date, search_terms)
        emmaa_stmts.append(es)
    return emmaa_stmts



import unittest

from .api import app


class EmmaaApiTest(unittest.TestCase):

    def setUp(self):
        app.testing = True
        self.app = app.test_client()

    def test_query_endpoint(self):
        """Test if we can submit a simple model query

        NOTE: only tests that the query was received by the API and that it
        was properly formatted.
        """

        test_query = {'user': {'email': 'joshua@emmaa.com',
                               'name': 'joshua',
                               'slack_id': '123456abcdef'},
                      'models': ['aml', 'luad'],
                      'query': {'objectSelection': 'ERK',
                                'subjectSelection': 'BRAF',
                                'typeSelection': 'activation'},
                      'register': 'false',
                      'test': 'true'}

        resp = self.app.post('/query/submit', json=test_query)

        assert resp.status_code == 200

    def test_query(self):
        """Test querying a model"""

        test_query = {'user': {'email': 'joshua@emmaa.com',
                                   'name': 'joshua',
                                   'slack_id': '123456abcdef'},
                          'models': ['rasmodel'],
                          'query': {'objectSelection': 'AKT1',
                                    'subjectSelection': 'PIK3CA',
                                    'typeSelection': 'activation'},
                          'register': 'false'}

        resp = self.app.post('/query/submit', json=test_query)

        assert resp.status_code == 200



"""The AWS Lambda emmaa-update-pipeline definition.

This file contains the function that starts model update cycle. It must be
placed on AWS Lambda, which can either be done manually (not recommended) or by
running:

$ python update_lambda.py update_pipeline.py emmaa-update-pipeline

in this directory.
"""

import boto3
import json


def lambda_handler(event, context):
    """Invoke individual model update functions.

    This function iterates through all models contained on S3 bucket and calls
    a different Lambda function to run model update for the models configured
    to be updated daily. It is expected that models have 'run_model_update'
    parameter in their config.json files.

    This function is designed to be placed on AWS Lambda, taking the event and
    context arguments that are passed. Note that this function must always have
    the same parameters, even if any or all of them are unused, because we do
    not have control over what Lambda sends as parameters. Parameters are
    unused in this function.

    Lambda is configured to automatically run this script every day.

    See the top of the page for the Lambda update procedure.

    Parameters
    ----------
    event : dict
        A dictionary containing metadata regarding the triggering event.
    context : object
        This is an object containing potentially useful context provided by
        Lambda. See the documentation cited above for details.

    Returns
    -------
    ret : dict
        A response returned by the latest call to emmaa-model-update function.
    """
    s3 = boto3.client('s3')
    lam = boto3.client('lambda')
    objs = s3.list_objects_v2(Bucket='emmaa', Prefix='models/', Delimiter='/')
    prefixes = objs['CommonPrefixes']
    for prefix_dict in prefixes:
        prefix = prefix_dict['Prefix']
        config_key = f'{prefix}config.json'
        obj = s3.get_object(Bucket='emmaa', Key=config_key)
        config = json.loads(obj['Body'].read().decode('utf8'))
        if config.get('run_daily_update', False):
            model_name = prefix[7:-1]
            payload = {"model": model_name}
            resp = lam.invoke(FunctionName='emmaa-model-update',
                              InvocationType='RequestResponse',
                              Payload=json.dumps(payload))
            print(resp['Payload'].read())
    return {'statusCode': 200, 'result': 'SUCCESS'}



import requests
from inflection import underscore
from collections import OrderedDict as _o
from .util import get_class_from_name
from indra.statements.statements import Statement, Agent, get_all_descendants,\
    mk_str, make_hash
from indra.databases.hgnc_client import get_hgnc_id
from indra.databases.chebi_client import get_chebi_id_from_name
from indra.databases.mesh_client import get_mesh_id_name
from indra.preassembler.grounding_mapper import gm


class Query(object):
    """The parent class of all query types."""
    @classmethod
    def _from_json(cls, json_dict):
        query_type = json_dict.get('type')
        query_cls = get_class_from_name(query_type, Query)
        query = query_cls._from_json(json_dict)
        return query

    def matches(self, other):
        return self.matches_key() == other.matches_key()

    def matches_key(self):
        pass

    def get_hash(self):
        return make_hash(self.matches_key(), 14)

    def get_hash_with_model(self, model_name):
        key = (self.matches_key(), model_name)
        return make_hash(mk_str(key), 14)


class StructuralProperty(Query):
    pass


class PathProperty(Query):
    """This type of query requires finding a mechanistic causally consistent
    path that satisfies query statement.

    Parameters
    ----------
    path_stmt : indra.statements.Statement
        A path to look for in the model represented as INDRA statement.
    entity_constraints : dict(list(indra.statements.Agent))
        A dictionary containing lists of Agents to be included in or excluded
        from the path.
    relationship_constraints : dict(list(str))
        A dictionary containing lists of Statement types to include in or
        exclude from the path.
    """
    def __init__(self, path_stmt, entity_constraints=None,
                 relationship_constraints=None):
        self.path_stmt = path_stmt
        if entity_constraints:
            self.include_entities = entity_constraints.get('include', [])
            self.exclude_entities = entity_constraints.get('exclude', [])
        else:
            self.include_entities = []
            self.exclude_entities = []
        if relationship_constraints:
            self.include_rels = relationship_constraints.get('include', [])
            self.exclude_rels = relationship_constraints.get('exclude', [])
        else:
            self.include_rels = []
            self.exclude_rels = []
        self.entities = self.get_entities()

    def to_json(self):
        query_type = underscore(type(self).__name__)
        json_dict = _o(type=query_type)
        json_dict['path'] = self.path_stmt.to_json()
        json_dict['entity_constraints'] = {}
        if self.include_entities:
            json_dict['entity_constraints']['include'] = [
                ec.to_json() for ec in self.include_entities]
        if self.exclude_entities:
            json_dict['entity_constraints']['exclude'] = [
                ec.to_json() for ec in self.exclude_entities]
        json_dict['relationship_constraints'] = {}
        if self.include_rels:
            json_dict['relationship_constraints']['include'] = [
                {'type': rel} for rel in self.include_rels]
        if self.exclude_rels:
            json_dict['relationship_constraints']['exclude'] = [
                {'type': rel} for rel in self.exclude_rels]
        return json_dict

    @classmethod
    def _from_json(cls, json_dict):
        path_stmt_json = json_dict.get('path')
        path_stmt = Statement._from_json(path_stmt_json)
        ent_constr_json = json_dict.get('entity_constraints')
        entity_constraints = None
        if ent_constr_json:
            entity_constraints = {}
            for key, value in ent_constr_json.items():
                entity_constraints[key] = [Agent._from_json(ec) for ec
                                           in value]
        rel_constr_json = json_dict.get('relationship_constraints')
        relationship_constraints = None
        if rel_constr_json:
            relationship_constraints = {}
            for key, value in rel_constr_json.items():
                relationship_constraints[key] = [
                    rel_type['type'] for rel_type in value]
        query = cls(path_stmt, entity_constraints, relationship_constraints)
        return query

    def get_entities(self):
        """Return entities from the path statement and the inclusion list."""
        path_entities = self.path_stmt.agent_list()
        return path_entities + self.include_entities

    def matches_key(self):
        key = self.path_stmt.matches_key()
        if self.include_entities:
            for ent in sorted(self.include_entities,
                              key=lambda x: x.matches_key()):
                key += ent.matches_key()
        if self.exclude_entities:
            for ent in sorted(self.exclude_entities,
                              key=lambda x: x.matches_key()):
                key += ent.matches_key()
        if self.include_rels:
            for rel in sorted(self.include_rels):
                key += rel
        if self.exclude_rels:
            for rel in sorted(self.exclude_rels):
                key += rel
        return mk_str(key)

    def __str__(self):
        parts = [f'PathPropertyQuery(stmt={str(self.path_stmt)}.']
        if self.include_entities:
            inents = ', '.join([str(e) for e in self.include_entities])
            parts.append(f' Include entities: {inents}.')
        if self.exclude_entities:
            exents = ', '.join([str(e) for e in self.exclude_entities])
            parts.append(f' Exclude entities: {exents}.')
        if self.include_rels:
            inrels = ', '.join(self.include_rels)
            parts.append(f' Include relations: {inrels}.')
        if self.exclude_rels:
            exrels = ', '.join(self.exclude_rels)
            parts.append(f' Exclude relations: {exrels}.')
        return ''.join(parts)

    def __repr__(self):
        return str(self)


class SimpleInterventionProperty(Query):
    pass


class ComparativeInterventionProperty(Query):
    pass


def get_agent_from_text(ag_name, use_grouding_service=True):
    """Return an INDRA Agent object."""
    grounding_url = "http://grounding.indra.bio/ground"
    if use_grouding_service:
        try:
            agent = get_agent_from_grounding_service(ag_name, grounding_url)
        except Exception as e:
            logger.warning('Could not get agent from grounding service: %s' % e)
            agent = get_agent_from_local_grounding(ag_name)
    else:
        agent = get_agent_from_local_grounding(ag_name)
    return agent


def get_grounding_from_name(name):
    """Return grounding given an agent name."""
    # See if it's a gene name
    hgnc_id = get_hgnc_id(name)
    if hgnc_id:
        return ('HGNC', hgnc_id)

    # Check if it's in the grounding map
    try:
        refs = gm[name]
        if isinstance(refs, dict):
            for dbn, dbi in refs.items():
                if dbn != 'TEXT':
                    return (dbn, dbi)
    # If not, search by text
    except KeyError:
        pass

    chebi_id = get_chebi_id_from_name(name)
    if chebi_id:
        return ('CHEBI', f'CHEBI:{chebi_id}')

    mesh_id, _ = get_mesh_id_name(name)
    if mesh_id:
        return ('MESH', mesh_id)

    return None


def get_agent_from_local_grounding(ag_name):
    grounding = get_grounding_from_name(ag_name)
    if not grounding:
        grounding = get_grounding_from_name(ag_name.upper())
        ag_name = ag_name.upper()
    if not grounding:
        raise GroundingError(f"Could not find grounding for {ag_name}.")
    agent = Agent(ag_name, db_refs={grounding[0]: grounding[1]})
    return agent


def get_agent_from_grounding_service(ag_name, url):
    res = requests.post(url, json={'text': ag_name})
    rj = res.json()
    if not rj:
        raise GroundingError(f"Could not find grounding for {ag_name}.")
    agent = Agent(name=rj[0]['term']['entry_name'],
                  db_refs={rj[0]['term']['db']: rj[0]['term']['id']})
    return agent


class GroundingError(Exception):
    pass



import json
import os
from os.path import abspath, dirname, join
from nose.plugins.attrib import attr
from indra.statements import Phosphorylation, Agent
from emmaa.queries import (Query, PathProperty, get_agent_from_local_grounding,
                           get_agent_from_grounding_service,
                           get_grounding_from_name, get_agent_from_text,)


def test_path_property_from_json():
    query_file = join(dirname(abspath(__file__)), 'path_property_query.json')
    with open(query_file, 'r') as f:
        json_dict = json.load(f)
    query = Query._from_json(json_dict)
    assert query
    assert isinstance(query, PathProperty)
    assert isinstance(query.path_stmt, Phosphorylation), query.path_stmt
    assert query.path_stmt.enz.name == 'EGFR', query.path_stmt
    assert query.path_stmt.sub.name == 'ERK', query.path_stmt
    assert isinstance(query.exclude_entities[0], Agent)
    assert query.exclude_entities[0].name == 'PI3K'
    assert isinstance(query.include_entities[0], Agent)
    assert query.include_entities[0].name == 'MAPK1'
    assert set(query.exclude_rels) == set(['IncreaseAmount', 'DecreaseAmount'])
    assert query.include_rels[0] == 'Inhibition'


def test_path_property_to_json():
    stmt = Phosphorylation(enz=Agent('EGFR', db_refs={'HGNC': '3236'}),
                           sub=Agent('ERK', db_refs={'FPLX': 'ERK'}))
    entity_constraints = {'exclude': [Agent('PI3K', db_refs={'FPLX': 'PI3K'})]}
    relationship_contraints = {'exclude': ['IncreaseAmount', 'DecreaseAmount']}
    query = PathProperty(stmt, entity_constraints, relationship_contraints)
    assert query
    json = query.to_json()
    assert json.get('type') == 'path_property'
    path = json.get('path')
    assert path.get('type') == 'Phosphorylation'
    deserialize_query = Query._from_json(json)
    json2 = deserialize_query.to_json()
    assert json == json2, {'json': json, 'json2': json2}


def test_stringify_path_property():
    stmt = Phosphorylation(enz=Agent('EGFR', db_refs={'HGNC': '3236'}),
                           sub=Agent('ERK', db_refs={'FPLX': 'ERK'}))
    entity_constraints = {'exclude': [Agent('PI3K', db_refs={'FPLX': 'PI3K'})]}
    relationship_contraints = {'exclude': ['IncreaseAmount', 'DecreaseAmount']}
    query = PathProperty(stmt, entity_constraints, relationship_contraints)
    query_str = str(query)
    assert query_str == 'PathPropertyQuery(stmt=Phosphorylation(EGFR(), ERK()). Exclude entities: PI3K(). Exclude relations: IncreaseAmount, DecreaseAmount.'


def test_grounding_from_name():
    assert get_grounding_from_name('MAPK1') == ('HGNC', '6871')
    assert get_grounding_from_name('BRAF') == ('HGNC', '1097')


def test_local_grounding():
    agent = get_agent_from_local_grounding('MAPK1')
    assert isinstance(agent, Agent)
    assert agent.name == 'MAPK1'
    assert agent.db_refs == {'HGNC': '6871'}
    # test with lower case
    agent = get_agent_from_local_grounding('mapk1')
    assert isinstance(agent, Agent)
    assert agent.name == 'MAPK1'
    assert agent.db_refs == {'HGNC': '6871'}
    # other agent
    agent = get_agent_from_local_grounding('BRAF')
    assert isinstance(agent, Agent)
    assert agent.name == 'BRAF'
    assert agent.db_refs == {'HGNC': '1097'}


@attr('nonpublic')
def test_grounding_service():
    agent = get_agent_from_text('MAPK1', use_grouding_service=True)
    assert isinstance(agent, Agent)
    assert agent.name == 'MAPK1'
    assert agent.db_refs == {'HGNC': '6871'}
    agent = get_agent_from_text('BRAF', use_grouding_service=True)
    assert isinstance(agent, Agent)
    assert agent.name == 'BRAF'
    assert agent.db_refs == {'HGNC': '1097'}



import unittest
from emmaa.statements import EmmaaStatement
from emmaa.priors import SearchTerm
from emmaa.readers.db_client_reader import read_db_pmid_search_terms


# FIXME Test should only run if tests are run locally, not by Travis
@unittest.skip('Test not run by Travis')
def test_read_db_pmid_search_terms():
    """Check read_db_pmid_search_terms() function with different inputs."""
    search_terms = [SearchTerm('gene', 'AKT2', {'HGNC': '392', 'UP': 'P31751'},
        'AKT2'), SearchTerm('gene', 'ACOX2', {'HGNC': '120', 'UP': 'Q99424'},
        'ACOX2')]
    # Check for empty input.
    assert len(read_db_pmid_search_terms({})) == 0
    # Check for PMIDs that do not have any statements.
    nostmts_pmid = "22178463"
    assert len(read_db_pmid_search_terms({nostmts_pmid: search_terms})) == 0
    # Check for PMIDs that have statements.
    stmts_pmid = "23431386"
    estmts = read_db_pmid_search_terms({stmts_pmid: search_terms})
    assert len(estmts) > 0
    assert isinstance(estmts[0], EmmaaStatement)
    estmts[0].search_terms == search_terms



import datetime
from indra_db.client.statements import get_statements_by_paper
from indra_db.util import get_primary_db
from emmaa.statements import EmmaaStatement


def read_db_pmid_search_terms(pmid_search_terms):
    """Return extracted EmmaaStatements from INDRA database given a
    PMID-search term dict.

    Parameters
    ----------
    pmid_search_terms : dict
        A dict representing a set of PMIDs pointing to search terms that
        produced them.

    Returns
    -------
    list[:py:class:`emmaa.model.EmmaaStatement`]
        A list of EmmaaStatements extracted from the given PMIDs.
    """
    pmids = list(pmid_search_terms.keys())
    date = datetime.datetime.utcnow()
    db = get_primary_db()
    pmid_stmts = get_statements_by_paper(pmids, id_type='pmid', db=db,
                                         preassembled=False)
    estmts = []
    for pmid, stmts in pmid_stmts.items():
        for stmt in stmts:
            es = EmmaaStatement(stmt, date, pmid_search_terms[pmid])
            estmts.append(es)
    return estmts



import boto3
import sys
from os import path
from zipfile import ZipFile

HERE = path.dirname(path.abspath(__file__))


def upload_function(script_name, function_name):
    """Upload the lambda function by pushing a zip file to Lambda.

    This function pre-supposes you are running from the same directory that
    contains the lambda script. 
    
    Parameters
    ----------
    script_name : str
        Name of a script containing lambda function. 
    function_name : object
        Name of a lambda function as specified on AWS Lambda.
    """
    lamb = boto3.client('lambda')
    with ZipFile(path.join(HERE, 'lambda.zip'), 'w') as zf:
        zf.write(path.join(HERE, script_name),
                 f'emmaa/{path.basename(HERE)}/{script_name}')
        zf.write(path.join(HERE, '__init__.py'),
                 f'emmaa/{path.basename(HERE)}/__init__.py')
        zf.write(path.join(HERE, path.pardir, '__init__.py'),
                 'emmaa/__init__.py')

    with open(path.join(HERE, 'lambda.zip'), 'rb') as zf:
        ret = lamb.update_function_code(ZipFile=zf.read(),
                                        FunctionName=function_name)
        print(ret)
    return

def main():
    script_name = sys.argv[1]
    function_name = sys.argv[2]
    upload_function(script_name, function_name)

if __name__ == '__main__':
    main()


import logging
from indra_db import client
from indra.tools import assemble_corpus as ac


logger = logging.getLogger(__name__)


def get_stmts_for_gene(gene):
    """Return all existing Statements for a given gene from the DB.

    Parameters
    ----------
    gene : str
        The HGNC symbol of a gene to query.

    Returns
    -------
    list[indra.statements.Statement]
        A list of INDRA Statements in which the given gene is involved.
    """
    return client.get_statements_by_gene_role_type(gene, preassembled=False,
                                                   count=100000)


def get_stmts_for_gene_list(gene_list, other_entities):
    """Return all Statements between genes in a given list.

    Parameters
    ----------
    gene_list : list[str]
        A list of HGNC symbols for genes to query.
    other_entities : list[str]
        A list of other entities to keep as part of the set of Statements.

    Returns
    -------
    list[indra.statements.Statement]
        A list of INDRA Statements between the given list of genes and other
        entities specified.
    """
    stmts = []
    for gene in gene_list:
        logger.info(f'Querying {gene}')
        st = get_stmts_for_gene(gene)
        logger.info(f'Got {len(st)} statements for {gene}')
        stmts += st
    stmts = ac.filter_gene_list(stmts, gene_list + other_entities, policy='all')
    return stmts



import os
import re
import boto3
import logging
from datetime import datetime
from botocore import UNSIGNED
from botocore.client import Config
from inflection import camelize
from indra.statements import get_all_descendants


FORMAT = '%Y-%m-%d-%H-%M-%S'
RE_DATEFORMAT = r'\d{4}\-\d{2}\-\d{2}\-\d{2}\-\d{2}\-\d{2}'
logger = logging.getLogger(__name__)


def strip_out_date(keystring):
    """Strips out datestring of format FORMAT from a keystring"""
    try:
        return re.search(RE_DATEFORMAT, keystring).group()
    except AttributeError:
        logger.warning(f'Can\'t parse string {keystring} for date')
        return None


def get_date_from_str(date_str):
    return datetime.strptime(date_str, FORMAT)


def make_date_str(date=None):
    """Return a date string in a standardized format.

    Parameters
    ----------
    date : Optional[datetime.datetime]
        A date object to get the standardized string for. If not provided,
        utcnow() is used to construct the date. (Note: using UTC is important
        because this code may run in multiple contexts).

    Returns
    -------
    str
        The datetime string in a standardized format.
    """
    if not date:
        date = datetime.utcnow()
    return date.strftime(FORMAT)


def sort_s3_files_by_date(bucket, prefix, extension=None):
    """
    Return the list of keys of the files on an S3 path sorted by date starting
    with the most recent one.
    """
    def process_key(key):
        fname_with_extension = os.path.basename(key)
        fname = os.path.splitext(fname_with_extension)[0]
        date_str = fname.split('_')[1]
        return get_date_from_str(date_str)
    client = get_s3_client()
    resp = client.list_objects(Bucket=bucket, Prefix=prefix)
    files = resp.get('Contents', [])
    if extension:
        files = [file for file in files if file['Key'].endswith(extension)]
    files = sorted(files, key=lambda f: process_key(f['Key']), reverse=True)
    return files


def find_latest_s3_file(bucket, prefix, extension=None):
    """Return the key of the file with latest date string on an S3 path"""
    files = sort_s3_files_by_date(bucket, prefix, extension)
    try:
        latest = files[0]['Key']
        return latest
    except IndexError:
        logger.info('File is not found.')


def find_second_latest_s3_file(bucket, prefix, extension=None):
    """Return the key of the file with second latest date string on an S3 path
    """
    files = sort_s3_files_by_date(bucket, prefix, extension)
    try:
        latest = files[1]['Key']
        return latest
    except IndexError:
        logger.info("File is not found.")


def find_latest_s3_files(number_of_files, bucket, prefix, extension=None):
    """
    Return the keys of the specified number of files with latest date strings
    on an S3 path sorted by date starting with the earliest one.
    """
    files = sort_s3_files_by_date(bucket, prefix, extension)
    keys = []
    for ix in range(number_of_files):
        keys.append(files[ix]['Key'])
    keys.reverse()
    return keys


def find_number_of_files_on_s3(bucket, prefix, extension=None):
    files = sort_s3_files_by_date(bucket, prefix, extension)
    return len(files)


def get_s3_client(unsigned=True):
    """Return a boto3 S3 client with optional unsigned config.

    Parameters
    ----------
    unsigned : Optional[bool]
        If True, the client will be using unsigned mode in which public
        resources can be accessed without credentials. Default: True

    Returns
    -------
    botocore.client.S3
        A client object to AWS S3.
    """
    if unsigned:
        return boto3.client('s3', config=Config(signature_version=UNSIGNED))
    else:
        return boto3.client('s3')


def get_class_from_name(cls_name, parent_cls):
    classes = get_all_descendants(parent_cls)
    for cl in classes:
        if cl.__name__.lower() == camelize(cls_name).lower():
            return cl
    raise NotAClassName(f'{cls_name} is not recognized as a '
                        f'{parent_cls.__name__} type!')


class NotAClassName(Exception):
    pass



"""This module implements the object model for EMMAA model testing."""
import json
import boto3
import pickle
import logging
import datetime
import itertools
import jsonpickle
from collections import defaultdict
from fnvhash import fnv1a_32
from indra.explanation.model_checker import PysbModelChecker, \
    PybelModelChecker, SignedGraphModelChecker, UnsignedGraphModelChecker
from indra.explanation.reporting import stmts_from_pysb_path, \
    stmts_from_pybel_path, stmts_from_indranet_path
from indra.assemblers.english.assembler import EnglishAssembler
from indra.sources.indra_db_rest.api import get_statement_queries
from indra.statements import Statement, Agent, Concept, Event
from indra.util.statement_presentation import group_and_sort_statements
from emmaa.model import EmmaaModel
from emmaa.util import make_date_str, get_s3_client, get_class_from_name
from emmaa.analyze_tests_results import TestRound, StatsGenerator
from emmaa.answer_queries import QueryManager


logger = logging.getLogger(__name__)


result_codes_link = 'https://emmaa.readthedocs.io/en/latest/dashboard/response_codes.html'
RESULT_CODES = {
    'STATEMENT_TYPE_NOT_HANDLED': 'Statement type not handled',
    'SUBJECT_MONOMERS_NOT_FOUND': 'Statement subject not in model',
    'SUBJECT_NOT_FOUND': 'Statement subject not in model',
    'OBSERVABLES_NOT_FOUND': 'Statement object state not in model',
    'OBJECT_NOT_FOUND': 'Statement object state not in model',
    'NO_PATHS_FOUND': 'No path found that satisfies the test statement',
    'MAX_PATH_LENGTH_EXCEEDED': 'Path found but exceeds search depth',
    'PATHS_FOUND': 'Path found which satisfies the test statement',
    'INPUT_RULES_NOT_FOUND': 'No rules with test statement subject',
    'MAX_PATHS_ZERO': 'Path found but not reconstructed',
    'QUERY_NOT_APPLICABLE': 'Query is not applicable for this model'
}
ARROW_DICT = {'Complex': u"\u2194",
              'Inhibition': u"\u22A3",
              'DecreaseAmount': u"\u22A3"}


class ModelManager(object):
    """Manager to generate and store properties of a model and relevant tests.

    Parameters
    ----------
    model : emmaa.model.EmmaaModel
        EMMAA model

    Attributes
    ----------
    mc_mapping : dict
        A dictionary mapping a ModelChecker type to a corresponding method
        for assembling the model and a ModelChecker class.
    mc_types : dict
        A dictionary in which each key is a type of a ModelChecker and value is
        a dictionary containing an instance of a model, an instance of a
        ModelChecker and a list of test results.
    entities : list[indra.statements.agent.Agent]
        A list of entities of EMMAA model.
    applicable_tests : list[emmaa.model_tests.EmmaaTest]
        A list of EMMAA tests applicable for given EMMAA model.
    make_links : bool
        Whether to include links to INDRA db in test results.
    """
    def __init__(self, model):
        self.model = model
        self.mc_mapping = {
            'pysb': (self.model.assemble_pysb, PysbModelChecker,
                     stmts_from_pysb_path),
            'pybel': (self.model.assemble_pybel, PybelModelChecker,
                      stmts_from_pybel_path),
            'signed_graph': (self.model.assemble_signed_graph,
                             SignedGraphModelChecker,
                             stmts_from_indranet_path),
            'unsigned_graph': (self.model.assemble_unsigned_graph,
                               UnsignedGraphModelChecker,
                               stmts_from_indranet_path)}
        self.mc_types = {}
        for mc_type in model.test_config.get('mc_types', ['pysb']):
            self.mc_types[mc_type] = {}
            assembled_model = self.mc_mapping[mc_type][0]()
            self.mc_types[mc_type]['model'] = assembled_model
            self.mc_types[mc_type]['model_checker'] = (
                self.mc_mapping[mc_type][1](assembled_model))
            self.mc_types[mc_type]['test_results'] = []
        self.entities = self.model.get_assembled_entities()
        self.applicable_tests = []
        self.make_links = model.test_config.get('make_links', True)

    def get_updated_mc(self, mc_type, stmts):
        """Update the ModelChecker and graph with stmts for tests/queries."""
        mc = self.mc_types[mc_type]['model_checker']
        mc.statements = stmts
        if mc_type == 'pysb':
            mc.graph = None
            mc.get_graph(prune_im=True, prune_im_degrade=True)
        return mc

    def add_test(self, test):
        """Add a test to a list of applicable tests."""
        self.applicable_tests.append(test)

    def add_result(self, mc_type, result):
        """Add a result to a list of results."""
        self.mc_types[mc_type]['test_results'].append(result)

    def run_all_tests(self):
        """Run all applicable tests with all available ModelCheckers."""
        max_path_length, max_paths = self._get_test_configs()
        for mc_type in self.mc_types:
            self.run_tests_per_mc(mc_type, max_path_length, max_paths)

    def run_tests_per_mc(self, mc_type, max_path_length, max_paths):
        """Run all applicable tests with one ModelChecker."""
        mc = self.get_updated_mc(
            mc_type, [test.stmt for test in self.applicable_tests])
        logger.info(f'Running the tests with {mc_type} ModelChecker.')
        results = mc.check_model(
            max_path_length=max_path_length, max_paths=max_paths)
        for (stmt, result) in results:
            self.add_result(mc_type, result)

    def make_english_path(self, mc_type, result):
        """Create an English description of a path."""
        paths = []
        if result.paths:
            for path in result.paths:
                sentences = []
                if mc_type == 'signed_graph' or mc_type == 'unsigned_graph':
                    for i in range(len(path[:-1])):
                        sentence = path[i][0] + ' -> ' + path[i+1][0]
                        sentences.append((sentence, ''))
                else:
                    if mc_type == 'pysb':
                        stmts = stmts_from_pysb_path(
                            path, self.mc_types['pysb']['model'],
                            self.model.assembled_stmts)
                    elif mc_type == 'pybel':
                        stmts = stmts_from_pybel_path(
                            path, self.mc_types['pybel']['model'],
                            from_db=False, stmts=self.model.assembled_stmts)
                    for stmt in stmts:
                        if not stmt:
                            continue
                        if isinstance(stmt, list):
                            stmt = stmt[0]
                        ea = EnglishAssembler([stmt])
                        sentence = ea.make_model()
                        if self.make_links:
                            link = get_statement_queries([stmt])[0] + '&format=html'
                            sentences.append((sentence, link))
                        else:
                            sentences.append((sentence, ''))
                paths.append(sentences)
        return paths

    def make_path_json(self, mc_type, result):
        paths = []
        if result.paths:
            for path in result.paths:
                path_nodes = []
                edge_list = []
                report_function = self.mc_mapping[mc_type][2]
                model = self.mc_types[mc_type]['model']
                stmts = self.model.assembled_stmts
                if mc_type == 'pysb':
                    report_stmts = report_function(path, model, stmts)
                    path_stmts = [[st] for st in report_stmts]
                    merge = False
                elif mc_type == 'pybel':
                    path_stmts = report_function(path, model, False, stmts)
                    merge = False
                elif mc_type == 'signed_graph':
                    path_stmts = report_function(path, model, True, False, stmts)
                    merge = True
                elif mc_type == 'unsigned_graph':
                    path_stmts = report_function(path, model, False, False, stmts)
                    merge = True
                for i, step in enumerate(path_stmts):
                    edge_nodes = []
                    if len(step) < 1:
                        continue
                    stmt_type = type(step[0]).__name__
                    for j, ag in enumerate(step[0].agent_list()):
                        if ag is not None:
                            edge_nodes.append(ag.name)
                        if j == (len(step[0].agent_list()) - 1):
                            break
                        if stmt_type in ARROW_DICT:
                            edge_nodes.append(ARROW_DICT[stmt_type])
                        else:
                            edge_nodes.append(u"\u2192")
                    if i == 0:
                        for n in edge_nodes:
                            path_nodes.append(n)
                    else:
                        for n in edge_nodes[1:]:
                            path_nodes.append(n)
                    step_sentences = self._make_path_stmts(step, merge=merge)
                    edge_dict = {'edge': ' '.join(edge_nodes),
                                 'stmts': step_sentences}
                    edge_list.append(edge_dict)
                path_json = {'path': ' '.join(path_nodes),
                             'edge_list': edge_list}
                paths.append(path_json)
        return paths

    def _make_path_stmts(self, stmts, merge=False):
        sentences = []
        if merge:
            groups = group_and_sort_statements(stmts)
            new_stmts = []
            for group in groups:
                stmt_type = group[0][-1]
                agent_names = group[0][1]
                if len(agent_names) != 2:
                    continue
                if stmt_type == 'Influence':
                    stmt = get_class_from_name(stmt_type, Statement)(
                        Event(Concept(agent_names[0])),
                        Event(Concept(agent_names[1])))
                else:
                    try:
                        stmt = get_class_from_name(stmt_type, Statement)(
                            Agent(agent_names[0]), Agent(agent_names[1]))
                    except ValueError:
                        stmt = get_class_from_name(stmt_type, Statement)(
                            [Agent(ag_name) for ag_name in agent_names])
                new_stmts.append(stmt)
            stmts = new_stmts
        for stmt in stmts:
            ea = EnglishAssembler([stmt])
            sentence = ea.make_model()
            if self.make_links:
                link = get_statement_queries([stmt])[0] + '&format=html'
                sentences.append((link, sentence))
            else:
                sentences.append(('', sentence))
        return sentences

    def make_english_result_code(self, result):
        """Get an English explanation of a result code."""
        result_code = result.result_code
        return [[(RESULT_CODES[result_code], result_codes_link)]]

    def make_result_code(self, result):
        result_code = result.result_code
        return RESULT_CODES[result_code]

    def answer_query(self, query):
        """Answer user query with a path if it is found."""
        if ScopeTestConnector.applicable(self, query):
            results = []
            for mc_type in self.mc_types:
                mc = self.get_updated_mc(mc_type, [query.path_stmt])
                max_path_length, max_paths = self._get_test_configs()
                result = mc.check_statement(
                    query.path_stmt, max_paths, max_path_length)
                results.append((mc_type, self.process_response(mc_type, result)))
            return results
        else:
            return [('', self.hash_response_list([[
                (RESULT_CODES['QUERY_NOT_APPLICABLE'], result_codes_link)]]))]

    def answer_queries(self, queries):
        """Answer all queries registered for this model.

        Parameters
        ----------
        queries : list[emmaa.queries.Query]
            A list of queries to run.

        Returns
        -------
        responses : list[tuple(json, json)]
            A list of tuples each containing a query, mc_type and result json.
        """
        responses = []
        applicable_queries = []
        applicable_stmts = []
        max_path_length, max_paths = self._get_test_configs()
        for query in queries:
            if ScopeTestConnector.applicable(self, query):
                applicable_queries.append(query)
                applicable_stmts.append(query.path_stmt)
            else:
                responses.append(
                    (query, '', self.hash_response_list(
                        [[(RESULT_CODES['QUERY_NOT_APPLICABLE'],
                          result_codes_link)]])))
        # Only do the following steps if there are applicable queries
        if applicable_queries:
            for mc_type in self.mc_types:
                mc = self.get_updated_mc(mc_type, applicable_stmts)
                results = mc.check_model()
                for ix, (_, result) in enumerate(results):
                    responses.append(
                        (applicable_queries[ix], mc_type,
                         self.process_response(mc_type, result)))
        return sorted(responses, key=lambda x: x[0].matches_key())

    def _get_test_configs(self):
        try:
            max_path_length = \
                self.model.test_config['statement_checking']['max_path_length']
        except KeyError:
            max_path_length = 5
        try:
            max_paths = \
                self.model.test_config['statement_checking']['max_paths']
        except KeyError:
            max_paths = 1
        logger.info('Parameters for model checking: %d, %d' %
                    (max_path_length, max_paths))
        return (max_path_length, max_paths)

    def process_response(self, mc_type, result):
        """Return a dictionary in which every key is a hash and value is a list
        of tuples. Each tuple contains a sentence describing either a step in a
        path (if it was found) or result code (if a path was not found) and a
        link leading to a webpage with more information about corresponding
        sentence.
        """
        if result.paths:
            response_list = self.make_english_path(mc_type, result)
        else:
            response_list = self.make_english_result_code(result)
        return self.hash_response_list(response_list)

    def hash_response_list(self, response_list):
        """Return a dictionary mapping a hash with a response in a response
        list.
        """
        response_dict = {}
        for response in response_list:
            sentences = []
            for sentence, _ in response:
                sentences.append(sentence)
            response_str = ' '.join(sentences)
            response_hash = str(fnv1a_32(response_str.encode('utf-8')))
            response_dict[response_hash] = response
        return response_dict

    def assembled_stmts_to_json(self):
        """Put assembled statements to JSON format."""
        stmts = []
        for stmt in self.model.assembled_stmts:
            stmts.append(stmt.to_json())
        return stmts

    def results_to_json(self):
        """Put test results to json format."""
        pickler = jsonpickle.pickler.Pickler()
        results_json = []
        results_json.append({
            'model_name': self.model.name,
            'statements': self.assembled_stmts_to_json(),
            'mc_types': [mc_type for mc_type in self.mc_types.keys()],
            'make_links': self.make_links})
        for ix, test in enumerate(self.applicable_tests):
            test_ix_results = {'test_type': test.__class__.__name__,
                               'test_json': test.to_json()}
            for mc_type in self.mc_types:
                result = self.mc_types[mc_type]['test_results'][ix]
                test_ix_results[mc_type] = {
                    'result_json': pickler.flatten(result),
                    'path_json': self.make_path_json(mc_type, result),
                    'result_code': self.make_result_code(result)}
            results_json.append(test_ix_results)
        return results_json


class TestManager(object):
    """Manager to generate and run a set of tests on a set of models.

    Parameters
    ----------
    model_managers : list[emmaa.model_tests.ModelManager]
        A list of ModelManager objects
    tests : list[emmaa.model_tests.EmmaaTest]
        A list of EMMAA tests
    """
    def __init__(self, model_managers, tests):
        self.model_managers = model_managers
        self.tests = tests

    def make_tests(self, test_connector):
        """Generate a list of applicable tests for each model with a given test
        connector.

        Parameters
        ----------
        test_connector : emmaa.model_tests.TestConnector
            A TestConnector object to use for connecting models to tests.
        """
        logger.info(f'Checking applicability of {len(self.tests)} tests to '
                    f'{len(self.model_managers)} models')
        for model_manager, test in itertools.product(self.model_managers,
                                                     self.tests):
            if test_connector.applicable(model_manager, test):
                model_manager.add_test(test)
                logger.debug(f'Test {test.stmt} is applicable')
            else:
                logger.debug(f'Test {test.stmt} is not applicable')
        logger.info(f'Created tests for {len(self.model_managers)} models.')
        for model_manager in self.model_managers:
            logger.info(f'Created {len(model_manager.applicable_tests)} tests '
                        f'for {model_manager.model.name} model.')

    def run_tests(self):
        """Run tests for a list of model-test pairs"""
        for model_manager in self.model_managers:
            model_manager.run_all_tests()


class TestConnector(object):
    """Determines if a given test is applicable to a given model."""
    def __init__(self):
        pass

    @staticmethod
    def applicable(model, test):
        """Return True if the test is applicable to the given model."""
        return True


class ScopeTestConnector(TestConnector):
    """Determines applicability of a test to a model by overlap in scope."""
    @staticmethod
    def applicable(model, test):
        """Return True of all test entities are in the set of model entities"""
        model_entities = model.entities
        test_entities = test.get_entities()
        # TODO
        # After adding entities as an attribute to StatementCheckingTest(), use
        # test_entities = test.entities
        return ScopeTestConnector._overlap(model_entities, test_entities)

    @staticmethod
    def _overlap(model_entities, test_entities):
        me_names = {e.name for e in model_entities}
        te_names = {e.name for e in test_entities}
        # If all test entities are in model entities, we get an empty set here
        # so we return True
        return not te_names - me_names


class EmmaaTest(object):
    """Represent an EMMAA test condition"""
    def get_entities(self):
        """Return a list of entities that the test checks for."""
        raise NotImplementedError()


class StatementCheckingTest(EmmaaTest):
    """Represent an EMMAA test condition that checks a PySB-assembled model
    against an INDRA Statement."""
    def __init__(self, stmt, configs=None):
        self.stmt = stmt
        self.configs = {} if not configs else configs
        logger.info('Test configs: %s' % configs)
        # TODO
        # Add entities as a property if we can reload tests on s3.
        # self.entities = self.get_entities()

    def check(self, model_checker, pysb_model):
        """Use a model checker to check if a given model satisfies the test."""
        max_path_length = self.configs.get('max_path_length', 5)
        max_paths = self.configs.get('max_paths', 1)
        logger.info('Parameters for model checking: %s, %d' %
                    (max_path_length, max_paths))
        res = model_checker.check_statement(
            self.stmt,
            max_path_length=max_path_length,
            max_paths=max_paths)
        return res

    def get_entities(self):
        """Return a list of entities that the test checks for."""
        return self.stmt.agent_list()

    def to_json(self):
        return self.stmt.to_json()

    def __repr__(self):
        return "%s(stmt=%s)" % (self.__class__.__name__, repr(self.stmt))


def load_tests_from_s3(test_name):
    """Load Emmaa Tests with the given name from S3.

    Parameters
    ----------
    test_name : str
        Looks for a test file in the emmaa bucket on S3 with key
        'tests/{test_name}'.

    Return
    ------
    list of EmmaaTest
        List of EmmaaTest objects loaded from S3.
    """
    client = get_s3_client()
    test_key = f'tests/{test_name}'
    logger.info(f'Loading tests from {test_key}')
    obj = client.get_object(Bucket='emmaa', Key=test_key)
    tests = pickle.loads(obj['Body'].read())
    return tests


def save_model_manager_to_s3(model_name, model_manager):
    client = get_s3_client(unsigned=False)
    logger.info(f'Saving a model manager for {model_name} model to S3.')
    client.put_object(Body=pickle.dumps(model_manager), Bucket='emmaa',
                      Key=f'results/{model_name}/latest_model_manager.pkl')


def run_model_tests_from_s3(model_name, upload_mm=True,
                            upload_results=True, upload_stats=True,
                            registered_queries=True, db=None):
    """Run a given set of tests on a given model, both loaded from S3.

    After loading both the model and the set of tests, model/test overlap
    is determined using a ScopeTestConnector and tests are run.


    Parameters
    ----------
    model_name : str
        Name of EmmaaModel to load from S3.
    upload_mm : Optional[bool]
        Whether to upload a model manager instance to S3 as a pickle file.
        Default: True
    upload_results : Optional[bool]
        Whether to upload test results to S3 in JSON format. Can be set
        to False when running tests. Default: True
    upload_stats : Optional[bool]
        Whether to upload latest statistics about model and a test.
        Default: True
    registered_queries : Optional[bool]
        If True, registered queries are fetched from the database and
        executed, the results are then saved to the database. Default: True
    db : Optional[emmaa.db.manager.EmmaaDatabaseManager]
        If given over-rides the default primary database.

    Returns
    -------
    emmaa.model_tests.ModelManager
        Instance of ModelManager containing the model data, list of applied
        tests and the test results.
    emmaa.analyze_test_results.StatsGenerator
        Instance of StatsGenerator containing statistics about model and test.
    """
    model = EmmaaModel.load_from_s3(model_name)
    test_corpus = model.test_config.get('test_corpus', 'large_corpus_tests.pkl')
    tests = load_tests_from_s3(test_corpus)
    mm = ModelManager(model)
    if upload_mm:
        save_model_manager_to_s3(model_name, mm)
    tm = TestManager([mm], tests)
    tm.make_tests(ScopeTestConnector())
    tm.run_tests()
    results_json_dict = mm.results_to_json()
    results_json_str = json.dumps(results_json_dict, indent=1)
    # Optionally upload test results to S3
    if upload_results:
        client = get_s3_client(unsigned=False)
        date_str = make_date_str()
        result_key = f'results/{model_name}/results_{date_str}.json'
        logger.info(f'Uploading test results to {result_key}')
        client.put_object(Bucket='emmaa', Key=result_key,
                          Body=results_json_str.encode('utf8'))
    tr = TestRound(results_json_dict)
    sg = StatsGenerator(model_name, latest_round=tr)
    sg.make_stats()

    # Optionally upload statistics to S3
    if upload_stats:
        sg.save_to_s3()
    if registered_queries:
        qm = QueryManager(db=db, model_managers=[mm])
        qm.answer_registered_queries(model_name)
    return (mm, sg)



import logging
import pickle
from datetime import datetime
from emmaa.util import get_s3_client, make_date_str
from emmaa.db import get_db


logger = logging.getLogger(__name__)


model_manager_cache = {}


class QueryManager(object):
    """Manager to run queries and interact with the database.

    Parameters
    ----------
    db : emmaa.db.EmmaaDatabaseManager
        An instance of a database manager to use.
    model_managers : list[emmaa.model_tests.ModelManager]
        Optional list of ModelManagers to use for running queries. If not
        given, the methods will load ModelManager from S3 when needed.
    """
    def __init__(self, db=None, model_managers=None):
        self.db = db
        if db is None:
            self.db = get_db('primary')
        self.model_managers = model_managers if model_managers else []

    def answer_immediate_query(
            self, user_email, user_id, query, model_names, subscribe):
        """This method first tries to find saved result to the query in the
        database and if not found, runs ModelManager method to answer query."""
        # Store query in the database for future reference.
        self.db.put_queries(user_email, user_id, query, model_names, subscribe)
        # Check if the query has already been answered for any of given models
        # and retrieve the results from database.
        saved_results = self.db.get_results_from_query(query, model_names)
        if not saved_results:
            saved_results = []
        checked_models = {res[0] for res in saved_results}
        # If the query was answered for all models before, return the results.
        if checked_models == set(model_names):
            return format_results(saved_results)
        # Run queries mechanism for models for which result was not found.
        new_results = []
        new_date = datetime.now()
        for model_name in model_names:
            if model_name not in checked_models:
                mm = self.get_model_manager(model_name)
                response_list = mm.answer_query(query)
                for (mc_type, response) in response_list:
                    new_results.append(
                        (model_name, query, mc_type, response, new_date))
                if subscribe:
                    self.db.put_results(
                        model_name,
                        [(query, mc_type, response)
                         for mc_type, response in response_list])
        all_results = saved_results + new_results
        return format_results(all_results)

    def answer_registered_queries(
            self, model_name, find_delta=True, notify=False):
        """Retrieve queries registered on database for a given model,
        answer them, calculate delta between results, notify users in case of
        any changes, and put results to a database.
        """
        model_manager = self.get_model_manager(model_name)
        queries = self.db.get_queries(model_name)
        # Only do the following steps if there are queries for this model
        if queries:
            results = model_manager.answer_queries(queries)
            new_results = [(model_name, result[0], result[1], result[2], '')
                           for result in results]
            # Optionally find delta between results
            # NOTE: For now the report is presented in the logs. In future we
            # can choose some other ways to keep track of result changes.
            if find_delta:
                reports = self.make_reports_from_results(new_results, False, 'str')
                for report in reports:
                    logger.info(report)
            self.db.put_results(model_name, results)

    def get_registered_queries(self, user_email):
        """Get formatted results to queries registered by user."""
        results = self.db.get_results(user_email)
        return format_results(results)

    def make_reports_from_results(
            self, new_results, stored=True, report_format='str'):
        """Make a report given latest results and queries the results are for.

        Parameters
        ----------
        new_results : list[tuple]
            Latest results as a list of tuples where each tuple has the format
            (model_name, query, mc_type, result_json, date).
        stored : bool
            Whether the new_results are already stored in the database.
        report_format : str
            A format to write reports in. Accepted values: 'str' and 'html'.
        Returns
        -------
        reports : list[str]
            A list of reports on changes for each of the queries.
        """
        processed_query_mc = []
        reports = []
        # If latest results are in db, retrieve the second latest
        if stored:
            order = 2
        # If latest results are not in db, retrieve the latest stored
        else:
            order = 1
        for model_name, query, mc_type, new_result_json, _ in new_results:
            if (model_name, query, mc_type) in processed_query_mc:
                continue
            try:
                old_results = self.db.get_results_from_query(
                    query, [model_name], order)
                if old_results:
                    for old_result in old_results:
                        if mc_type == old_result[2]:
                            old_result_json = old_result[3]
                            if report_format == 'str':
                                report = self.make_str_report_one_query(
                                    model_name, query, mc_type,
                                    new_result_json, old_result_json)
                            elif report_format == 'html':
                                report = self._make_html_one_query_inner(
                                    model_name, query, mc_type,
                                    new_result_json, old_result_json)
                            reports.append(report)
                else:
                    logger.info('No previous result was found.')
                    if report_format == 'str':
                        report = self.make_str_report_one_query(
                            model_name, query, mc_type, new_result_json, None)
                    elif report_format == 'html':
                        report = self._make_html_one_query_inner(
                            model_name, query, mc_type, new_result_json, None)
                    reports.append(report)
            except IndexError:
                logger.info('No result for desired date back was found.')
                if report_format == 'str':
                    report = self.make_str_report_one_query(
                        model_name, query, mc_type, new_result_json, None)
                elif report_format == 'html':
                    report = self._make_html_one_query_inner(
                        model_name, query, mc_type, new_result_json, None)
                reports.append(report)
            processed_query_mc.append((model_name, query, mc_type))
        return reports

    def get_user_query_delta(
            self, user_email, filename='query_delta', report_format='str'):
        """Produce a report for all query results per user in a given format."""
        results = self.db.get_results(user_email, latest_order=1)
        if report_format == 'str':
            filename = filename + '.txt'
            self.make_str_report_per_user(results, filename=filename)
        elif report_format == 'html':
            filename = filename + '.html'
            self.make_html_report_per_user(results, filename=filename)

    def get_report_per_query(self, model_name, query):
        try:
            new_results = self.db.get_results_from_query(
                            query, [model_name], latest_order=1)
        except IndexError:
            logger.info('No latest result was found.')
            return None
        return self.make_reports_from_results(new_results, True, 'str')

    def make_str_report_per_user(self, results, filename='query_delta.txt'):
        """Produce a report for all query results per user in a text file."""
        reports = self.make_reports_from_results(results, True, 'str')
        with open(filename, 'w') as f:
            for report in reports:
                f.write(report)

    def make_html_report_per_user(self, results, filename='query_delta.html'):
        """Produce a report for all query results per user in an html file."""
        msg = '<html><body>'
        reports = self.make_reports_from_results(results, True, 'html')
        for report in reports:
            msg += report
        msg += '</body></html>'
        with open(filename, 'w') as f:
            f.write(msg)

    def make_str_report_one_query(self, model_name, query, mc_type,
                                  new_result_json, old_result_json=None):
        """Return a string message containing information about a query and any
        change in the results."""
        if is_query_result_diff(new_result_json, old_result_json):
            if not old_result_json:
                msg = f'\nThis is the first result to query {query} ' \
                      f'in {model_name} with {mc_type} model checker.' \
                      f'\nThe result is:'
                msg += _process_result_to_str(new_result_json)
            else:
                msg = f'\nA new result to query {query} in {model_name} was ' \
                      f'found with {mc_type} model checker. '
                msg += '\nPrevious result was:'
                msg += _process_result_to_str(old_result_json)
                msg += '\nNew result is:'
                msg += _process_result_to_str(new_result_json)
        else:
            msg = f'\nA result to query {query} in ' \
                  f'{model_name} from {mc_type} model checker ' \
                  f'did not change. The result is:'
            msg += _process_result_to_str(new_result_json)
        return msg

    def make_html_one_query_report(self, model_name, query, mc_type,
                                   new_result_json, old_result_json=None):
        """Return an html page containing information about a query and any
        change in the results."""
        msg = '<html><body>'
        msg += self._make_html_one_query_inner(
            model_name, query, mc_type, new_result_json, old_result_json)
        msg += '</body></html>'
        return msg

    def _make_html_one_query_inner(self, model_name, query, mc_type,
                                   new_result_json, old_result_json=None):
        # Create an html part for one query to be used in producing html report
            if is_query_result_diff(new_result_json, old_result_json):
                if not old_result_json:
                    msg = f'<p>This is the first result to query {query} in ' \
                          f'{model_name} with {mc_type} model checker. ' \
                          f'The result is:<br>'
                    msg += _process_result_to_html(new_result_json)
                    msg += '</p>'
                else:
                    msg = f'<p>A new result to query {query} in ' \
                          f'{model_name} was found with {mc_type} ' \
                          f'model checker.<br>'
                    msg += '<br>Previous result was:<br>'
                    msg += _process_result_to_html(old_result_json)
                    msg += '<br>New result is:<br>'
                    msg += _process_result_to_html(new_result_json)
                    msg += '</p>'
            else:
                msg = f'<p>A result to query {query} in ' \
                      f'{model_name} from {mc_type} model checker did not ' \
                      f'change. The result is:<br>'
                msg += _process_result_to_html(new_result_json)
                msg += '</p>'
            return msg

    def notify_user(
            self, user_email, model_name, query, mc_type, new_result_json,
            old_result_json=None):
        """Create a query result delta report and send it to user."""
        str_msg = self.make_str_report_one_query(
            model_name, query, mc_type, new_result_json, old_result_json)
        html_msg = self.make_html_one_query_report(
            model_name, query, mc_type, new_result_json, old_result_json)
        # TODO send an email to user
        pass

    def get_model_manager(self, model_name):
        # Try get model manager from class attributes or load from s3.
        for mm in self.model_managers:
            if mm.model.name == model_name:
                return mm
        return load_model_manager_from_s3(model_name)

    def _recreate_db(self):
        self.db.drop_tables(force=True)
        self.db.create_tables()


def is_query_result_diff(new_result_json, old_result_json=None):
    """Return True if there is a delta between results."""
    # NOTE: this function is query-type specific so it may need to be
    # refactored as a method of the Query class:

    # Return True if this is the first result
    if not old_result_json:
        return True
    # Compare hashes of query results
    old_result_hashes = [k for k in old_result_json.keys()]
    new_result_hashes = [k for k in new_result_json.keys()]
    return not set(new_result_hashes) == set(old_result_hashes)


def format_results(results):
    """Format db output to a standard json structure."""
    formatted_results = []
    for result in results:
        formatted_result = {}
        formatted_result['model'] = result[0]
        query = result[1]
        formatted_result['query'] = _make_query_simple_dict(query)
        formatted_result['mc_type'] = result[2]
        response_json = result[3]
        response = _process_result_to_html(response_json)
        formatted_result['response'] = response
        formatted_result['date'] = make_date_str(result[4])
        formatted_results.append(formatted_result)
    return formatted_results


def load_model_manager_from_s3(model_name):
    model_manager = model_manager_cache.get(model_name)
    if model_manager:
        logger.info(f'Loaded model manager for {model_name} from cache.')
        return model_manager
    client = get_s3_client()
    key = f'results/{model_name}/latest_model_manager.pkl'
    logger.info(f'Loading latest model manager for {model_name} model from '
                f'S3.')
    obj = client.get_object(Bucket='emmaa', Key=key)
    body = obj['Body'].read()
    model_manager = pickle.loads(body)
    model_manager_cache[model_name] = model_manager
    return model_manager


def _process_result_to_str(result_json):
    # Remove the links when making text report
    msg = '\n'
    for v in result_json.values():
        for sentence, link in v:
            msg += sentence
    return msg


def _process_result_to_html(result_json):
    # Make clickable links when making htmk report
    response_list = []
    for v in result_json.values():
        for ix, (sentence, link) in enumerate(v):
            if ix > 0:
                response_list.append('<br>')
            if link:
                response_list.append(
                    f'<a href="{link}" target="_blank" '
                    f'class="status-link">{sentence}</a>')
            else:
                response_list.append(f'<a>{sentence}</a>')
        response = ''.join(response_list)
    return response


def _make_query_simple_dict(query):
    """Turn Query object into a simple dictionary for easier representation on
    the dashboard."""
    query_dict = {}
    stmt = query.path_stmt
    query_dict['typeSelection'] = type(stmt).__name__
    subj, obj = stmt.agent_list()
    query_dict['subjectSelection'] = subj.name
    query_dict['objectSelection'] = obj.name
    return query_dict



import datetime
from indra.statements import Activation, ActivityCondition, Phosphorylation, \
    Agent, Evidence
from emmaa.model import EmmaaModel
from emmaa.priors import SearchTerm
from emmaa.statements import EmmaaStatement


def test_model_extend():
    ev1 = Evidence(pmid='1234', text='abcd', source_api='x')
    ev2 = Evidence(pmid='1234', text='abcde', source_api='x')
    ev3 = Evidence(pmid='1234', text='abcd', source_api='x')
    indra_sts = [Phosphorylation(None, Agent('a'), evidence=ev) for ev in
                 [ev1, ev2, ev3]]
    emmaa_sts = [EmmaaStatement(st, datetime.datetime.now(), ['x']) for st in
                 indra_sts]
    em = EmmaaModel('x', {'search_terms': [], 'ndex': {'network': None}})
    em.add_statements([emmaa_sts[0]])
    em.extend_unique(emmaa_sts[1:])
    assert len(em.stmts) == 2
    stmt = EmmaaStatement(Phosphorylation(None, Agent('b'), evidence=ev1),
                          datetime.datetime.now(), ['x'])
    em.extend_unique([stmt])
    assert len(em.stmts) == 3


def test_model_json():
    """Test the json structure and content of EmmaaModel.to_json() output"""
    indra_stmts = \
        [Activation(Agent('BRAF', db_refs={'HGNC': '20974'}),
                    Agent('MAP2K1'),
                    evidence=[Evidence(text='BRAF activates MAP2K1.')]),
         Activation(Agent('MAP2K1',
                          activity=ActivityCondition('activity', True)),
                    Agent('MAPK1'),
                    evidence=[Evidence(text='Active MAP2K1 activates MAPK1.')])
         ]
    st = SearchTerm('gene', 'MAP2K1', db_refs={}, search_term='MAP2K1')
    emmaa_stmts = [EmmaaStatement(stmt, datetime.datetime.now(), [st])
                    for stmt in indra_stmts]
    config_dict = {'ndex': {'network': 'a08479d1-24ce-11e9-bb6a-0ac135e8bacf'},
                   'search_terms': [{'db_refs': {'HGNC': '20974'},
                                     'name': 'MAPK1',
                                     'search_term': 'MAPK1',
                                     'type': 'gene'}]}
    emmaa_model = EmmaaModel('test', config_dict)
    emmaa_model.add_statements(emmaa_stmts)

    emmaa_model_json = emmaa_model.to_json()

    # Test json structure
    assert emmaa_model_json['name'] == 'test'
    assert isinstance(emmaa_model_json['stmts'], list)
    assert emmaa_model_json['ndex_network'] == \
        'a08479d1-24ce-11e9-bb6a-0ac135e8bacf'

    # Test config
    assert emmaa_model_json['search_terms'][0]['type'] == 'gene'
    assert emmaa_model_json['search_terms'][0]['db_refs'] == {'HGNC': '20974'}

    # Test json statements
    assert 'BRAF activates MAP2K1.' == \
           emmaa_model_json['stmts'][0]['stmt']['evidence'][0]['text']
    assert 'BRAF activates MAP2K1.' == \
           emmaa_model_json['stmts'][0]['stmt']['evidence'][0]['text']
    assert 'Active MAP2K1 activates MAPK1.' == \
           emmaa_model_json['stmts'][1]['stmt']['evidence'][0]['text']
    assert emmaa_model_json['stmts'][0]['stmt']['subj']['name'] == 'BRAF'
    assert emmaa_model_json['stmts'][1]['stmt']['subj']['name'] == 'MAP2K1'
    assert emmaa_model_json['stmts'][1]['stmt']['obj']['name'] == 'MAPK1'

    # Need hashes to be strings so that javascript can read them
    assert isinstance(emmaa_model_json['stmts'][0]['stmt']['evidence'][0][
                          'source_hash'], str)


def test_filter_relevance():
    config_dict = {'ndex': {'network': 'a08479d1-24ce-11e9-bb6a-0ac135e8bacf'},
                   'search_terms': [{'db_refs': {'HGNC': '20974'},
                                     'name': 'MAPK1',
                                     'search_term': 'MAPK1',
                                     'type': 'gene'}]}
    indra_stmts = \
        [Activation(Agent('BRAF', db_refs={'HGNC': '20974'}),
                    Agent('MAP2K1'),
                    evidence=[Evidence(text='BRAF activates MAP2K1.',
                                       source_api='assertion')]),
         Activation(Agent('MAP2K1',
                          activity=ActivityCondition('activity', True)),
                    Agent('MAPK1'),
                    evidence=[Evidence(text='Active MAP2K1 activates '
                                            'MAPK1.',
                                       source_api='assertion')])
         ]
    st = SearchTerm('gene', 'MAP2K1', db_refs={}, search_term='MAP2K1')
    emmaa_stmts = [EmmaaStatement(stmt, datetime.datetime.now(), [st])
                   for stmt in indra_stmts]

    # Try no filter first
    emmaa_model = EmmaaModel('test', config_dict)
    emmaa_model.extend_unique(emmaa_stmts)
    emmaa_model.run_assembly()
    assert len(emmaa_model.assembled_stmts) == 2, emmaa_model.assembled_stmts

    # Next do a prior_one filter
    config_dict['assembly'] = {'filter_relevance': 'prior_one'}
    emmaa_model = EmmaaModel('test', config_dict)
    emmaa_model.extend_unique(emmaa_stmts)
    emmaa_model.run_assembly()
    assert len(emmaa_model.assembled_stmts) == 1, emmaa_model.assembled_stmts
    assert emmaa_model.assembled_stmts[0].obj.name == 'MAPK1'

    # Next do a prior_all filter
    config_dict['assembly'] = {'filter_relevance': 'prior_all'}
    emmaa_model = EmmaaModel('test', config_dict)
    emmaa_model.extend_unique(emmaa_stmts)
    emmaa_model.run_assembly()
    assert len(emmaa_model.assembled_stmts) == 0



from emmaa.priors import SearchTerm
from emmaa.priors.gene_list_prior import GeneListPrior


def test_get_search_terms():
    gp = GeneListPrior(['BRAF'], 'braf', 'BRAF model')
    assert gp.name == 'braf'
    assert gp.human_readable_name == 'BRAF model'
    st = gp.make_search_terms()
    assert st
    assert all([isinstance(s, SearchTerm) for s in st])
    assert st[0].type == 'gene'
    assert st[0].search_term == '"BRAF"'
    assert st[1].type == 'drug'



import json
import logging
import jsonpickle
import datetime
from collections import defaultdict
from emmaa.util import (find_latest_s3_file, find_second_latest_s3_file,
                        find_latest_s3_files, find_number_of_files_on_s3,
                        make_date_str, get_s3_client)
from indra.statements.statements import Statement
from indra.assemblers.english.assembler import EnglishAssembler
from indra.sources.indra_db_rest.api import get_statement_queries


logger = logging.getLogger(__name__)


CONTENT_TYPE_FUNCTION_MAPPING = {
    'statements': 'get_stmt_hashes',
    'applied_tests': 'get_applied_test_hashes',
    'passed_tests': 'get_passed_test_hashes',
    'paths': 'get_passed_test_hashes'}


class TestRound(object):
    """Analyzes the results of one test round.

    Parameters
    ----------
    json_results : list[dict]
        A list of JSON formatted dictionaries to store information about the
        test results. The first dictionary contains information about the
        model. Each consecutive dictionary contains information about a single
        test applied to the model and test results.

    Attributes
    ----------
    statements : list[indra.statements.Statement]
        A list of INDRA Statements used to assemble a model.
    mc_types_results : dict
        A dictionary mapping a type of a ModelChecker to a list of test
        results generated by this ModelChecker
    tests : list[indra.statements.Statement]
        A list of INDRA Statements used to make EMMAA tests.
    function_mapping : dict
        A dictionary of strings mapping a type of content to a tuple of
        functions necessary to find delta for this type of content. First
        function in a tuple gets a list of all hashes for a given content type,
        while the second returns an English description of a given content type
        for a single hash.
    """
    def __init__(self, json_results):
        self.json_results = json_results
        self.statements = self._get_statements()
        mc_types = self.json_results[0].get('mc_types', ['pysb'])
        self.mc_types_results = {}
        for mc_type in mc_types:
            self.mc_types_results[mc_type] = self._get_results(mc_type)
        self.make_links = self.json_results[0].get('make_links', True)
        self.tests = self._get_tests()
        self.function_mapping = CONTENT_TYPE_FUNCTION_MAPPING
        self.english_test_results = self._get_applied_tests_results()

    @classmethod
    def load_from_s3_key(cls, key):
        client = get_s3_client()
        logger.info(f'Loading test results from {key}')
        obj = client.get_object(Bucket='emmaa', Key=key)
        json_results = json.loads(obj['Body'].read().decode('utf8'))
        test_round = TestRound(json_results)
        return test_round

    # Model Summary Methods
    def get_total_statements(self):
        """Return a total number of statements in a model."""
        total = len(self.statements)
        logger.info(f'An assembled model has {total} statements.')
        return total

    def get_stmt_hashes(self):
        """Return a list of hashes for all statements in a model."""
        return [str(stmt.get_hash(refresh=True)) for stmt in self.statements]

    def get_statement_types(self):
        """Return a sorted list of tuples containing a statement type and a
        number of times a statement of this type occured in a model.
        """
        statement_types = defaultdict(int)
        logger.info('Finding a distribution of statements types.')
        for stmt in self.statements:
            statement_types[type(stmt).__name__] += 1
        return sorted(statement_types.items(), key=lambda x: x[1], reverse=True)

    def get_agent_distribution(self):
        """Return a sorted list of tuples containing an agent name and a number
        of times this agent occured in statements of a model."""
        logger.info('Finding agent distribution among model statements.')
        agent_count = defaultdict(int)
        for stmt in self.statements:
            for agent in stmt.agent_list():
                if agent is not None:
                    agent_count[agent.name] += 1
        return sorted(agent_count.items(), key=lambda x: x[1], reverse=True)

    def get_statements_by_evidence(self):
        """Return a sorted list of tuples containing a statement hash and a
        number of times this statement occured in a model."""
        stmts_evidence = {}
        for stmt in self.statements:
            stmts_evidence[str(stmt.get_hash(refresh=True))] = len(stmt.evidence)
        logger.info('Sorting statements by evidence count.')
        return sorted(stmts_evidence.items(), key=lambda x: x[1], reverse=True)

    def get_english_statements_by_hash(self):
        """Return a dictionary mapping a statement and its English description."""
        stmts_by_hash = {}
        for stmt in self.statements:
            stmts_by_hash[str(stmt.get_hash(refresh=True))] = (
                self.get_english_statement(stmt))
        return stmts_by_hash

    def get_english_statement(self, stmt):
        ea = EnglishAssembler([stmt])
        sentence = ea.make_model()
        if self.make_links:
            link = get_statement_queries([stmt])[0] + '&format=html'
        else:
            link = ''
        return (link, sentence)

    # Test Summary Methods
    def get_applied_test_hashes(self):
        """Return a list of hashes for all applied tests."""
        return list(self.english_test_results.keys())

    def get_passed_test_hashes(self, mc_type='pysb'):
        """Return a list of hashes for passed tests."""
        return [test_hash for test_hash in self.english_test_results.keys() if
                self.english_test_results[test_hash][mc_type][0] == 'Pass']

    def get_total_applied_tests(self):
        """Return a number of all applied tests."""
        total = len(self.tests)
        logger.info(f'{total} tests were applied.')
        return total

    def get_number_passed_tests(self, mc_type='pysb'):
        """Return a number of all passed tests."""
        total = len(self.get_passed_test_hashes(mc_type))
        logger.info(f'{total} tests passed.')
        return total

    def passed_over_total(self, mc_type='pysb'):
        """Return a ratio of passed over total tests."""
        return self.get_number_passed_tests(mc_type)/self.get_total_applied_tests()

    def _get_applied_tests_results(self):
        """Return a dictionary mapping a test hash and a list containing its
        English description, result in Pass/Fail form and either a path if it
        was found or a result code if it was not."""
        tests_by_hash = {}
        logger.info('Retrieving test hashes, english tests and test results.')

        def get_pass_fail(res):
            # Here use result.path_found because we care if the path was found
            # and do not care about path length
            if res.path_found:
                return 'Pass'
            else:
                return 'Fail'

        for ix, test in enumerate(self.tests):
            test_hash = str(test.get_hash(refresh=True))
            tests_by_hash[test_hash] = {
                'test': self.get_english_statement(test)}
            for mc_type in self.mc_types_results:
                result = self.mc_types_results[mc_type][ix]
                tests_by_hash[test_hash][mc_type] = [
                        get_pass_fail(result),
                        self.get_path_or_code_by_hash(test_hash, mc_type)]
        return tests_by_hash

    def get_path_descriptions(self, mc_type='pysb'):
        paths_by_test = {}
        results = self.mc_types_results[mc_type]
        for ix, result in enumerate(results):
            if result.paths:
                try:
                    paths = self.json_results[ix+1][mc_type]['path_json']
                except KeyError:
                    paths = []
                paths_by_test[str(self.tests[ix].get_hash(refresh=True))] = paths
        return paths_by_test

    def get_english_codes(self, mc_type):
        english_codes = {}
        results = self.mc_types_results[mc_type]
        for ix, result in enumerate(results):
            try:
                code = self.json_results[ix+1][mc_type]['result_code']
            except KeyError:
                code = result.result_code
            english_codes[str(self.tests[ix].get_hash(refresh=True))] = code
        return english_codes

    def get_pass_fail_by_hash(self, test_hash, mc_type='pysb'):
        return self.english_test_results[test_hash][mc_type][0]

    def get_path_or_code_by_hash(self, test_hash, mc_type='pysb'):
        # If we have a path description return it, otherwise return code
        try:
            result = self.get_path_descriptions(mc_type)[test_hash]
        except KeyError:
            result = self.get_english_codes(mc_type)[test_hash]
        return result

    # Methods to find delta
    def find_numeric_delta(self, other_round, one_round_numeric_func,
                           mc_type='pysb'):
        """Find a numeric delta between two rounds using a passed function.

        Parameters
        ----------
        other_round : emmaa.analyze_tests_results.TestRound
            A different instance of a TestRound
        one_round_numeric_function : str
            A name of a method to calculate delta. Accepted values:
            - get_total_statements
            - get_total_applied_tests
            - get_number_passed_tests
            - passed_over_total
        mc_type : str
            A name of a ModelChecker used to run tests.

        Returns
        -------
        delta : int or float
            Difference between return values of one_round_numeric_function
            of two given test rounds.
        """
        logger.info(f'Calculating numeric delta using {one_round_numeric_func}'
                    ' method')
        if one_round_numeric_func == 'get_number_passed_tests' or \
                one_round_numeric_func == 'passed_over_total':
            kwargs = {'mc_type': mc_type}
        else:
            kwargs = {}
        delta = (getattr(self, one_round_numeric_func)(**kwargs)
                 - getattr(other_round, one_round_numeric_func)(**kwargs))
        return delta

    def find_delta_hashes(self, other_round, content_type, **kwargs):
        """Return a dictionary of changed hashes of a given content type. This
        method makes use of self.function_mapping dictionary.

        Parameters
        ----------
        other_round : emmaa.analyze_tests_results.TestRound
            A different instance of a TestRound
        content_type : str
            A type of the content to find delta. Accepted values:
            - statements
            - applied_tests
            - passed_tests
            - paths
        **kwargs : dict
            For some of content types, additional arguments must be
            provided sych as mc_type.
        Returns
        -------
        hashes : dict
            A dictionary containing lists of added and removed hashes of a
            given content type between two test rounds.
        """
        logger.info(f'Finding a hashes delta for {content_type}.')
        latest_hashes = getattr(
            self, self.function_mapping[content_type])(**kwargs)
        logger.info(f'Found {len(latest_hashes)} hashes in current round.')
        previous_hashes = getattr(
            other_round,
            other_round.function_mapping[content_type])(**kwargs)
        logger.info(f'Found {len(previous_hashes)} hashes in other round.')
        # Find hashes unique for each of the rounds - this is delta
        added_hashes = list(set(latest_hashes) - set(previous_hashes))
        removed_hashes = list(set(previous_hashes) - set(latest_hashes))
        hashes = {'added': added_hashes, 'removed': removed_hashes}
        return hashes

    # Helping methods
    def _get_statements(self):
        serialized_stmts = self.json_results[0]['statements']
        return [Statement._from_json(stmt) for stmt in serialized_stmts]

    def _get_results(self, mc_type):
        unpickler = jsonpickle.unpickler.Unpickler()
        test_results = [unpickler.restore(result[mc_type]['result_json'])
                        for result in self.json_results[1:]]
        return test_results

    def _get_tests(self):
        tests = [Statement._from_json(res['test_json'])
                 for res in self.json_results[1:]]
        return tests


class StatsGenerator(object):
    """Generates statistic for a given test round.

    Parameters
    ----------
    model_name : str
        A name of a model the tests were run against.
    latest_round : emmaa.analyze_tests_results.TestRound
        An instance of a TestRound to generate statistics for. If not given,
        will be generated by loading test results from s3.
    previous_round : emmaa.analyze_tests_results.TestRound
        A different instance of a TestRound to find delta between two rounds.
        If not given, will be generated by loading test results from s3.

    Attributes
    ----------
    json_stats : dict
        A JSON-formatted dictionary containing test model and test statistics.
    previous_json_stats : list[dict]
        A JSON-formatted dictionary containing test model and test
        statistics for previous test round.
    """

    def __init__(self, model_name, latest_round=None, previous_round=None,
                 previous_json_stats=None):
        self.model_name = model_name
        if not latest_round:
            self.latest_round = self._get_latest_round()
        else:
            self.latest_round = latest_round
        if not previous_round:
            self.previous_round = self._get_previous_round()
        else:
            self.previous_round = previous_round
        self.json_stats = {}
        if not previous_json_stats:
            self.previous_json_stats = self._get_previous_json_stats()
        else:
            self.previous_json_stats = previous_json_stats

    def make_stats(self):
        """Check if two latest test rounds were found and add statistics to
        json_stats dictionary. If both latest round and previous round
        were passed or found on s3, a dictionary will have four key-value
        pairs: model_summary, test_round_summary, model_delta, and tests_delta.
        """
        if not self.latest_round:
            logger.info(f'Latest round for {self.model_name} is not found.')
            return
        logger.info(f'Generating stats for {self.model_name}.')
        self.make_model_summary()
        self.make_test_summary()
        self.make_model_delta()
        self.make_tests_delta()
        self.make_changes_over_time()

    def make_model_summary(self):
        """Add latest model state summary to json_stats."""
        logger.info(f'Generating model summary for {self.model_name}.')
        self.json_stats['model_summary'] = {
            'model_name': self.model_name,
            'number_of_statements': self.latest_round.get_total_statements(),
            'stmts_type_distr': self.latest_round.get_statement_types(),
            'agent_distr': self.latest_round.get_agent_distribution(),
            'stmts_by_evidence': self.latest_round.get_statements_by_evidence(),
            'all_stmts': self.latest_round.get_english_statements_by_hash()
        }

    def make_test_summary(self):
        """Add latest test round summary to json_stats."""
        logger.info(f'Generating test summary for {self.model_name}.')
        self.json_stats['test_round_summary'] = {
            'number_applied_tests': self.latest_round.get_total_applied_tests(),
            'all_test_results': self.latest_round.english_test_results}
        for mc_type in self.latest_round.mc_types_results:
            self.json_stats['test_round_summary'][mc_type] = {
                'number_passed_tests': (
                    self.latest_round.get_number_passed_tests(mc_type)),
                'passed_ratio': self.latest_round.passed_over_total(mc_type)}

    def make_model_delta(self):
        """Add model delta between two latest model states to json_stats."""
        logger.info(f'Generating model delta for {self.model_name}.')
        if not self.previous_round:
            self.json_stats['model_delta'] = {
                'statements_hashes_delta': {'added': [], 'removed': []}}
        else:
            self.json_stats['model_delta'] = {
                'statements_hashes_delta': self.latest_round.find_delta_hashes(
                    self.previous_round, 'statements')}

    def make_tests_delta(self):
        """Add tests delta between two latest test rounds to json_stats."""
        logger.info(f'Generating tests delta for {self.model_name}.')
        if not self.previous_round:
            tests_delta = {
                'applied_hashes_delta': {'added': [], 'removed': []}}
        else:
            tests_delta = {
                'applied_hashes_delta': self.latest_round.find_delta_hashes(
                    self.previous_round, 'applied_tests')}

        for mc_type in self.latest_round.mc_types_results:
            if not self.previous_round or mc_type not in \
                    self.previous_round.mc_types_results:
                tests_delta[mc_type] = {
                    'passed_hashes_delta': {'added': [], 'removed': []}}
            else:
                tests_delta[mc_type] = {
                    'passed_hashes_delta': self.latest_round.find_delta_hashes(
                        self.previous_round, 'passed_tests', mc_type=mc_type)}
        self.json_stats['tests_delta'] = tests_delta

    def make_changes_over_time(self):
        """Add changes to model and tests over time to json_stats."""
        logger.info(f'Comparing changes over time for {self.model_name}.')
        self.json_stats['changes_over_time'] = {
            'number_of_statements': self.get_over_time(
                'model_summary', 'number_of_statements'),
            'number_applied_tests': self.get_over_time(
                'test_round_summary', 'number_applied_tests'),
            'dates': self.get_dates()}
        for mc_type in self.latest_round.mc_types_results:
            self.json_stats['changes_over_time'][mc_type] = {
                'number_passed_tests': self.get_over_time(
                    'test_round_summary', 'number_passed_tests', mc_type),
                'passed_ratio': self.get_over_time(
                    'test_round_summary', 'passed_ratio', mc_type)}

    def get_over_time(self, section, metrics, mc_type='pysb'):
        logger.info(f'Getting changes over time in {metrics} '
                    f'for {self.model_name}.')
        # Not mc_type relevant data
        if metrics == 'number_of_statements' or \
                metrics == 'number_applied_tests':
            # First available stats
            if not self.previous_json_stats:
                previous_data = []
            else:
                previous_data = (
                    self.previous_json_stats['changes_over_time'][metrics])
            previous_data.append(self.json_stats[section][metrics])
        # Mc_type relevant data
        else:
            # First available stats
            if not self.previous_json_stats:
                previous_data = []
            else:
                # This mc_type wasn't available in previous stats
                if mc_type not in \
                        self.previous_json_stats['changes_over_time']:
                    previous_data = []
                else:
                    previous_data = (
                        self.previous_json_stats[
                            'changes_over_time'][mc_type][metrics])
            previous_data.append(self.json_stats[section][mc_type][metrics])
        return previous_data

    def get_dates(self):
        if not self.previous_json_stats:
            previous_dates = []
        else:
            previous_dates = (
                self.previous_json_stats['changes_over_time']['dates'])
        previous_dates.append(make_date_str())
        return previous_dates

    def save_to_s3(self):
        json_stats_str = json.dumps(self.json_stats, indent=1)
        client = get_s3_client(unsigned=False)
        date_str = make_date_str()
        stats_key = f'stats/{self.model_name}/stats_{date_str}.json'
        logger.info(f'Uploading test round statistics to {stats_key}')
        client.put_object(Bucket='emmaa', Key=stats_key,
                          Body=json_stats_str.encode('utf8'))

    def _get_latest_round(self):
        latest_key = find_latest_s3_file(
            'emmaa', f'results/{self.model_name}/results_', extension='.json')
        if latest_key is None:
            logger.info(f'Could not find a key to the latest test results '
                        f'for {self.model_name} model.')
            return
        tr = TestRound.load_from_s3_key(latest_key)
        return tr

    def _get_previous_round(self):
        previous_key = find_second_latest_s3_file(
            'emmaa', f'results/{self.model_name}/results_', extension='.json')
        if previous_key is None:
            logger.info(f'Could not find a key to the previous test results '
                        f'for {self.model_name} model.')
            return
        tr = TestRound.load_from_s3_key(previous_key)
        return tr

    def _get_previous_json_stats(self):
        key = find_latest_s3_file(
            'emmaa', f'stats/{self.model_name}/stats_', extension='.json')
        if key is None:
            logger.info(f'Could not find a key to the previous statistics '
                        f'for {self.model_name} model.')
            return
        client = get_s3_client()
        logger.info(f'Loading earlier statistics from {key}')
        obj = client.get_object(Bucket='emmaa', Key=key)
        previous_json_stats = json.loads(obj['Body'].read().decode('utf8'))
        return previous_json_stats



import json
import time
import pickle
import logging
import datetime
from indra.databases import ndex_client
import indra.tools.assemble_corpus as ac
from indra.literature import pubmed_client, elsevier_client
from indra.assemblers.cx import CxAssembler
from indra.assemblers.pysb import PysbAssembler
from indra.assemblers.pybel import PybelAssembler
from indra.assemblers.indranet import IndraNetAssembler
from indra.mechlinker import MechLinker
from indra.preassembler.hierarchy_manager import get_wm_hierarchies
from indra.preassembler import Preassembler
from indra.belief.wm_scorer import get_eidos_scorer
from indra.statements import Event, Association
from emmaa.priors import SearchTerm
from emmaa.readers.aws_reader import read_pmid_search_terms
from emmaa.readers.db_client_reader import read_db_pmid_search_terms
from emmaa.readers.elsevier_eidos_reader import \
    read_elsevier_eidos_search_terms
from emmaa.util import make_date_str, find_latest_s3_file, get_s3_client


logger = logging.getLogger(__name__)


class EmmaaModel(object):
    """"Represents an EMMAA model.

    Parameters
    ----------
    name : str
        The name of the model.
    config : dict
        A configuration dict that is typically loaded from a YAML file.

    Attributes
    ----------
    name : str
        A string containing the name of the model
    stmts : list[emmaa.EmmaaStatement]
        A list of EmmaaStatement objects representing the model
    assembly_config : dict
        Configurations for assembling the model.
    test_config : dict
        Configurations for running tests on the model.
    reading_config : dict
        Configurations for reading the content.
    search_terms : list[emmaa.priors.SearchTerm]
        A list of SearchTerm objects containing the search terms used in the
        model.
    ndex_network : str
        The identifier of the NDEx network corresponding to the model.
    assembled_stmts : list[indra.statements.Statement]
        A list of assembled INDRA Statements
    """
    def __init__(self, name, config):
        self.name = name
        self.stmts = []
        self.assembly_config = {}
        self.test_config = {}
        self.reading_config = {}
        self.search_terms = []
        self.ndex_network = None
        self._load_config(config)
        self.assembled_stmts = []

    def add_statements(self, stmts):
        """"Add a set of EMMAA Statements to the model

        Parameters
        ----------
        stmts : list[emmaa.EmmaaStatement]
            A list of EMMAA Statements to add to the model
        """
        self.stmts += stmts

    def get_indra_stmts(self):
        """Return the INDRA Statements contained in the model.

        Returns
        -------
        list[indra.statements.Statement]
            The list of INDRA Statements that are extracted from the EMMAA
            Statements.
        """
        return [es.stmt for es in self.stmts]

    def _load_config(self, config):
        self.search_terms = [SearchTerm.from_json(s) for s in
                             config['search_terms']]
        if 'ndex' in config:
            self.ndex_network = config['ndex']['network']
        else:
            self.ndex_network = None
        if 'reading' in config:
            self.reading_config = config['reading']
        if 'assembly' in config:
            self.assembly_config = config['assembly']
        if 'test' in config:
            self.test_config = config['test']

    def search_literature(self, date_limit=None):
        """Search for the model's search terms in the literature.

        Parameters
        ----------
        date_limit : Optional[int]
            The number of days to search back from today.

        Returns
        -------
        ids_to_terms : dict
            A dict representing all the literature source IDs (e.g.,
            PMIDs or PIIS) returned by the searches as keys,
            and the search terms for which the given ID was produced as
            values.
        """
        lit_source = self.reading_config.get('literature_source', 'pubmed')
        if lit_source == 'pubmed':
            terms_to_ids = self.search_pubmed(self.search_terms, date_limit)
        elif lit_source == 'elsevier':
            terms_to_ids = self.search_elsevier(self.search_terms, date_limit)
        else:
            raise ValueError('Unknown literature source: %s' % lit_source)
        ids_to_terms = {}
        for term, ids in terms_to_ids.items():
            for id in ids:
                try:
                    ids_to_terms[id].append(term)
                except KeyError:
                    ids_to_terms[id] = [term]
        return ids_to_terms

    @staticmethod
    def search_pubmed(search_terms, date_limit):
        """Search PubMed for given search terms.

        Parameters
        ----------
        search_terms : list[emmaa.priors.SearchTerm]
            A list of SearchTerm objects to search PubMed for.
        date_limit : int
            The number of days to search back from today.

        Returns
        -------
        terms_to_pmids : dict
            A dict representing given search terms as keys and PMIDs returned
            by searches as values.
        """
        terms_to_pmids = {}
        for term in search_terms:
            pmids = pubmed_client.get_ids(term.search_term, reldate=date_limit)
            logger.info(f'{len(pmids)} PMIDs found for {term.search_term}')
            terms_to_pmids[term] = pmids
            time.sleep(1)
        return terms_to_pmids

    @staticmethod
    def search_elsevier(search_terms, date_limit):
        """Search Elsevier for given search terms.

        Parameters
        ----------
        search_terms : list[emmaa.priors.SearchTerm]
            A list of SearchTerm objects to search PubMed for.
        date_limit : int
            The number of days to search back from today.

        Returns
        -------
        terms_to_piis : dict
            A dict representing given search terms as keys and PIIs returned
            by searches as values.
        """
        start_date = (
            datetime.datetime.utcnow() - datetime.timedelta(days=date_limit))
        start_date = start_date.isoformat(timespec='seconds') + 'Z'
        terms_to_piis = {}
        for term in search_terms:
            # NOTE for now limiting the search to only 5 PIIs
            piis = elsevier_client.get_piis_for_date(
                term.search_term, loaded_after=start_date)[:5]
            logger.info(f'{len(piis)} PIIs found for {term.search_term}')
            terms_to_piis[term] = piis
            time.sleep(1)
        return terms_to_piis

    def get_new_readings(self, date_limit=10):
        """Search new literature, read, and add to model statements"""
        reader = self.reading_config.get('reader', 'indra_db')
        ids_to_terms = self.search_literature(date_limit=date_limit)
        if reader == 'aws':
            estmts = read_pmid_search_terms(ids_to_terms)
        elif reader == 'indra_db':
            estmts = read_db_pmid_search_terms(ids_to_terms)
        elif reader == 'elsevier_eidos':
            estmts = read_elsevier_eidos_search_terms(ids_to_terms)
        else:
            raise ValueError('Unknown reader: %s' % reader)

        self.extend_unique(estmts)

    def extend_unique(self, estmts):
        """Extend model statements only if it is not already there."""
        source_hashes = {est.stmt.get_hash(shallow=False, refresh=True)
                         for est in self.stmts}
        for estmt in estmts:
            if estmt.stmt.get_hash(shallow=False, refresh=True) not in \
                    source_hashes:
                self.stmts.append(estmt)

    def eliminate_copies(self):
        """Filter out exact copies of the same Statement."""
        logger.info('Starting with %d raw EmmaaStatements' % len(self.stmts))
        self.stmts = list({estmt.stmt.get_hash(shallow=False, refresh=True):
                           estmt for estmt in self.stmts}.values())
        logger.info(('Continuing with %d raw EmmaaStatements'
                     ' that are not exact copies') % len(self.stmts))

    def run_assembly(self):
        """Run INDRA's assembly pipeline on the Statements."""
        self.eliminate_copies()
        stmts = self.get_indra_stmts()
        stmts = self.filter_event_association(stmts)
        stmts = ac.filter_no_hypothesis(stmts)
        if not self.assembly_config.get('skip_map_grounding'):
            stmts = ac.map_grounding(stmts)
        if self.assembly_config.get('standardize_names'):
            ac.standardize_names_groundings(stmts)
        if self.assembly_config.get('filter_ungrounded'):
            score_threshold = self.assembly_config.get('score_threshold')
            stmts = ac.filter_grounded_only(
                stmts, score_threshold=score_threshold)
        if self.assembly_config.get('merge_groundings'):
            stmts = ac.merge_groundings(stmts)
        if self.assembly_config.get('merge_deltas'):
            stmts = ac.merge_deltas(stmts)
        relevance_policy = self.assembly_config.get('filter_relevance')
        if relevance_policy:
            stmts = self.filter_relevance(stmts, relevance_policy)
        if not self.assembly_config.get('skip_filter_human'):
            stmts = ac.filter_human_only(stmts)
        if not self.assembly_config.get('skip_map_sequence'):
            stmts = ac.map_sequence(stmts)
        # Use WM hierarchies and belief scorer for WM preassembly
        preassembly_mode = self.assembly_config.get('preassembly_mode')
        if preassembly_mode == 'wm':
            hierarchies = get_wm_hierarchies()
            belief_scorer = get_eidos_scorer()
            stmts = ac.run_preassembly(
                stmts, return_toplevel=False, belief_scorer=belief_scorer,
                hierarchies=hierarchies)
        else:
            stmts = ac.run_preassembly(stmts, return_toplevel=False)
        belief_cutoff = self.assembly_config.get('belief_cutoff')
        if belief_cutoff is not None:
            stmts = ac.filter_belief(stmts, belief_cutoff)
        stmts = ac.filter_top_level(stmts)

        if self.assembly_config.get('filter_direct'):
            stmts = ac.filter_direct(stmts)
            stmts = ac.filter_enzyme_kinase(stmts)
            stmts = ac.filter_mod_nokinase(stmts)
            stmts = ac.filter_transcription_factor(stmts)

        if self.assembly_config.get('mechanism_linking'):
            ml = MechLinker(stmts)
            ml.gather_explicit_activities()
            ml.reduce_activities()
            ml.gather_modifications()
            ml.reduce_modifications()
            ml.gather_explicit_activities()
            ml.replace_activations()
            ml.require_active_forms()
            stmts = ml.statements

        self.assembled_stmts = stmts

    def filter_event_association(self, stmts):
        """Filter a list of Statements to exclude Events and Associations."""
        logger.info('Filtering Events and Associations.')
        stmts = [stmt for stmt in stmts if (
            (not isinstance(stmt, Event)) and
            (not isinstance(stmt, Association)))]
        logger.info('%d statements after filter...' % len(stmts))
        return stmts

    def filter_relevance(self, stmts, policy=None):
        """Filter a list of Statements to ones matching a search term."""
        logger.info('Filtering %d statements for relevance...' % len(stmts))
        stmts_out = []
        stnames = {s.name for s in self.search_terms}
        for stmt in stmts:
            agnames = {a.name for a in stmt.agent_list() if a is not None}
            if policy == 'prior_one' and (agnames & stnames):
                stmts_out.append(stmt)
            elif policy == 'prior_all' and agnames.issubset(stnames):
                stmts_out.append(stmt)
            elif policy is None:
                stmts_out.append(stmt)
        logger.info('%d statements after filter...' % len(stmts_out))
        return stmts_out

    def update_to_ndex(self):
        """Update assembled model as CX on NDEx, updates existing network."""
        if not self.assembled_stmts:
            self.run_assembly()
        cxa = CxAssembler(self.assembled_stmts, network_name=self.name)
        cxa.make_model()
        cx_str = cxa.print_cx()
        ndex_client.update_network(cx_str, self.ndex_network)

    def upload_to_ndex(self):
        """Upload the assembled model as CX to NDEx, creates new network."""
        if not self.assembled_stmts:
            self.run_assembly()
        cxa = CxAssembler(self.assembled_stmts, network_name=self.name)
        cxa.make_model()
        model_uuid = cxa.upload_model()
        self.ndex_network = model_uuid
        return model_uuid

    def save_to_s3(self):
        """Dump the model state to S3."""
        date_str = make_date_str()
        fname = f'models/{self.name}/model_{date_str}'
        client = get_s3_client(unsigned=False)
        # Dump as pickle
        client.put_object(Body=pickle.dumps(self.stmts), Bucket='emmaa',
                          Key=fname+'.pkl')
        # Dump as json
        client.put_object(Body=str.encode(json.dumps(self.to_json()),
                                          encoding='utf8'),
                          Bucket='emmaa', Key=fname+'.json')

    @classmethod
    def load_from_s3(klass, model_name):
        """Load the latest model state from S3.

        Parameters
        ----------
        model_name : str
            Name of model to load. This function expects the latest model
            to be found on S3 in the emmaa bucket with key
            'models/{model_name}/model_{date_string}', and the model config
            file at 'models/{model_name}/config.json'.

        Returns
        -------
        emmaa.model.EmmaaModel
            Latest instance of EmmaaModel with the given name, loaded from S3.
        """
        config = load_config_from_s3(model_name)
        stmts = load_stmts_from_s3(model_name)
        em = klass(model_name, config)
        em.stmts = stmts
        return em

    def get_entities(self):
        """Return a list of Agent objects that the model contains."""
        istmts = self.get_indra_stmts()
        agents = []
        for stmt in istmts:
            agents += [a for a in stmt.agent_list() if a is not None]
        return agents

    def get_assembled_entities(self):
        """Return a list of Agent objects that the assembled model contains."""
        if not self.assembled_stmts:
            self.run_assembly()
        agents = []
        for stmt in self.assembled_stmts:
            agents += [a for a in stmt.agent_list() if a is not None]
        return agents

    def assemble_pysb(self):
        """Assemble the model into PySB and return the assembled model."""
        if not self.assembled_stmts:
            self.run_assembly()
        pa = PysbAssembler()
        pa.add_statements(self.assembled_stmts)
        pysb_model = pa.make_model()
        return pysb_model

    def assemble_pybel(self):
        """Assemble the model into PyBEL and return the assembled model."""
        if not self.assembled_stmts:
            self.run_assembly()
        pba = PybelAssembler(self.assembled_stmts)
        pybel_model = pba.make_model()
        return pybel_model

    def assemble_signed_graph(self):
        """Assemble the model into signed graph and return the assembled graph."""
        if not self.assembled_stmts:
            self.run_assembly()
        ia = IndraNetAssembler(self.assembled_stmts)
        signed_graph = ia.make_model(graph_type='signed')
        return signed_graph

    def assemble_unsigned_graph(self):
        """Assemble the model into unsigned graph and return the assembled graph."""
        if not self.assembled_stmts:
            self.run_assembly()
        ia = IndraNetAssembler(self.assembled_stmts)
        unsigned_graph = ia.make_model(graph_type='digraph')
        return unsigned_graph

    def to_json(self):
        """Convert the model into a json dumpable dictionary"""
        json_output = {'name': self.name,
                       'ndex_network': self.ndex_network,
                       'search_terms': [st.to_json() for st
                                        in self.search_terms],
                       'stmts': [st.to_json() for st in self.stmts]}
        return json_output

    def __repr__(self):
        return "EmmaModel(%s, %d stmts, %d search terms)" % \
                   (self.name, len(self.stmts), len(self.search_terms))


def load_config_from_s3(model_name):
    """Return a JSON dict of config settings for a model from S3.

    Parameters
    ----------
    model_name : str
        The name of the model whose config should be loaded.

    Returns
    -------
    config : dict
        A JSON dictionary of the model configuration loaded from S3.
    """
    client = get_s3_client()
    base_key = f'models/{model_name}'
    config_key = f'{base_key}/config.json'
    logger.info(f'Loading model config from {config_key}')
    obj = client.get_object(Bucket='emmaa', Key=config_key)
    config = json.loads(obj['Body'].read().decode('utf8'))
    return config


def save_config_to_s3(model_name, config):
    """Upload config settings for a model to S3.

    Parameters
    ----------
    model_name : str
        The name of the model whose config should be saved to S3.
    config : dict
        A JSON dict of configurations for the model.
    """
    client = get_s3_client(unsigned=False)
    base_key = f'models/{model_name}'
    config_key = f'{base_key}/config.json'
    logger.info(f'Saving model config to {config_key}')
    client.put_object(Body=str.encode(json.dumps(config, indent=1),
                                      encoding='utf8'),
                      Bucket='emmaa', Key=config_key)


def load_stmts_from_s3(model_name):
    """Return the list of EMMAA Statements constituting the latest model.

    Parameters
    ----------
    model_name : str
        The name of the model whose config should be loaded.

    Returns
    -------
    stmts : list of emmaa.statements.EmmaaStatement
        The list of EMMAA Statements in the latest model version.
    """
    client = get_s3_client()
    base_key = f'models/{model_name}'
    latest_model_key = find_latest_s3_file('emmaa', f'{base_key}/model_',
                                           extension='.pkl')
    logger.info(f'Loading model state from {latest_model_key}')
    obj = client.get_object(Bucket='emmaa', Key=latest_model_key)
    stmts = pickle.loads(obj['Body'].read())
    return stmts



import os
import json
from nose.plugins.attrib import attr
from emmaa.analyze_tests_results import TestRound, StatsGenerator


TestRound.__test__ = False


path_here = os.path.abspath(os.path.dirname(__file__))
previous_results_file = os.path.join(path_here, 'previous_results.json')
new_results_file = os.path.join(path_here, 'new_results.json')
previous_stats_file = os.path.join(path_here, 'previous_stats.json')
with open(previous_results_file, 'r') as f:
    previous_results = json.load(f)
with open(new_results_file, 'r') as f:
    new_results = json.load(f)
with open(previous_stats_file, 'r') as f:
    previous_stats = json.load(f)


@attr('nonpublic')
def test_test_round():
    tr = TestRound(previous_results)
    assert tr
    assert tr.get_total_statements() == 2
    assert len(tr.get_stmt_hashes()) == 2
    assert tr.get_statement_types() == [('Activation', 2)]
    assert all(agent_tuple in tr.get_agent_distribution() for agent_tuple in
               [('BRAF', 1), ('MAP2K1', 2), ('MAPK1', 1)])
    assert all((stmt_hash, 1) in tr.get_statements_by_evidence() for stmt_hash
               in tr.get_stmt_hashes())
    assert tr.get_total_applied_tests() == 1
    assert tr.get_number_passed_tests() == 1
    assert tr.get_applied_test_hashes() == tr.get_passed_test_hashes()
    assert tr.passed_over_total() == 1.0
    tr2 = TestRound(new_results)
    assert tr2
    assert tr2.find_numeric_delta(tr, 'get_total_statements') == 2
    assert tr2.find_numeric_delta(tr, 'get_total_applied_tests') == 1
    assert tr2.find_numeric_delta(tr, 'get_number_passed_tests') == 1
    assert tr2.find_numeric_delta(tr, 'passed_over_total') == 0


@attr('nonpublic')
def test_stats_generator():
    latest_round = TestRound(new_results)
    previous_round = TestRound(previous_results)
    sg = StatsGenerator('test', latest_round=latest_round,
                        previous_round=previous_round,
                        previous_json_stats=previous_stats)
    sg.make_stats()
    assert sg.json_stats
    model_summary = sg.json_stats['model_summary']
    assert model_summary['number_of_statements'] == 4
    assert model_summary['stmts_type_distr'] == [('Activation', 4)]
    assert all(agent_tuple in model_summary['agent_distr'] for
               agent_tuple in [('AKT', 2), ('BRAF', 2), ('MAP2K1', 2),
                               ('MTOR', 1), ('MAPK1', 1)])
    assert len(model_summary['stmts_by_evidence']) == 4
    assert len(model_summary['all_stmts']) == 4
    test_round_summary = sg.json_stats['test_round_summary']
    assert test_round_summary['number_applied_tests'] == 2
    assert len(test_round_summary['all_test_results']) == 2
    assert test_round_summary['pysb']['number_passed_tests'] == 2
    assert test_round_summary['pysb']['passed_ratio'] == 1.0
    model_delta = sg.json_stats['model_delta']
    tests_delta = sg.json_stats['tests_delta']
    assert len(tests_delta['applied_hashes_delta']['added']) == 1
    assert len(tests_delta['pysb']['passed_hashes_delta']['added']) == 1
    changes = sg.json_stats['changes_over_time']
    assert changes['number_of_statements'] == [2, 4]
    assert changes['number_applied_tests'] == [1, 2]
    assert len(changes['dates']) == 2
    assert changes['pysb']['number_passed_tests'] == [1, 2]
    assert changes['pysb']['passed_ratio'] == [1, 1]



from nose.plugins.attrib import attr
from indra.explanation.model_checker import PathResult, PysbModelChecker
from indra.statements.statements import Statement
from emmaa.model import EmmaaModel
from emmaa.model_tests import (StatementCheckingTest, run_model_tests_from_s3,
                               load_tests_from_s3, ModelManager)
from emmaa.analyze_tests_results import TestRound, StatsGenerator

from emmaa.tests.test_db import _get_test_db


# Tell nose to not run tests in the imported modules
StatementCheckingTest.__test__ = False
run_model_tests_from_s3.__test__ = False
load_tests_from_s3.__test__ = False
TestRound.__test__ = False


def test_load_tests_from_s3():
    tests = load_tests_from_s3('simple_model_test.pkl')
    assert isinstance(tests, list)
    assert len(tests) == 1
    test = tests[0]
    assert isinstance(test, StatementCheckingTest)


@attr('nonpublic')
def test_run_tests_from_s3():
    db = _get_test_db()
    (mm, sg) = run_model_tests_from_s3(
        'test', upload_mm=False, upload_results=False, upload_stats=False,
        registered_queries=False, db=db)
    assert isinstance(mm, ModelManager)
    assert isinstance(mm.model, EmmaaModel)
    assert isinstance(mm.mc_types['pysb']['model_checker'], PysbModelChecker)
    assert isinstance(mm.mc_types['pysb']['test_results'][0], PathResult)
    assert isinstance(mm.applicable_tests[0], StatementCheckingTest)
    assert len(mm.applicable_tests) == 1
    assert len(mm.mc_types['pysb']['test_results']) == 1
    assert isinstance(sg, StatsGenerator)
    assert isinstance(sg.latest_round, TestRound)
    assert isinstance(sg.latest_round.mc_types_results['pysb'][0], PathResult)
    assert isinstance(sg.latest_round.statements[0], Statement)
    assert len(sg.latest_round.statements) == 2
    assert len(sg.latest_round.mc_types_results['pysb']) == 1
    assert len(sg.latest_round.tests) == 1



import os
import csv
import json
import logging
import numpy as np
import networkx as nx
from scipy.sparse.linalg import expm_multiply
from indra.util import batch_iter
from indra.sources import tas
from indra.databases import cbio_client, uniprot_client
from indra.databases.hgnc_client import hgnc_ids, get_hgnc_id, get_hgnc_name, \
                                        get_uniprot_id
from emmaa.priors import get_drugs_for_gene, SearchTerm


logger = logging.getLogger(__name__)


class TcgaCancerPrior(object):
    """Prior network generation using TCGA mutations for a given cancer type.

    This class implements building a prior network using a generic underlying
    prior, and TCGA data for a specific cancer type. Mutations for the given
    cancer type are extracted from TCGA studies and heat diffusion from the
    corresponding nodes in the prior is used to identify a set of relevant
    nodes.
    """
    def __init__(self, tcga_study_prefix, sif_prior, diffusion_service=None,
                 mutation_cache=None):
        if tcga_study_prefix not in tcga_studies:
            raise ValueError('TCGA study prefix must be one of %s' %
                             (', '.join(tcga_studies.keys())))
        # e.g. paad
        self.tcga_study_prefix = tcga_study_prefix
        self.mutations = None
        self.norm_mutations = None
        self.prior_graph = None
        self.sif_prior = sif_prior
        if diffusion_service is None:
            self.diffusion_service = 'http://v3.heat-diffusion.cytoscape.io:80'
        else:
            self.diffusion_service = diffusion_service
        self.mutation_cache = mutation_cache

    def make_prior(self, pct_heat_threshold=99):
        """Run the prior node list generation and return relevant nodes."""
        self.get_mutated_genes()
        self.load_sif_prior(self.sif_prior)
        res = self.get_relevant_nodes(pct_heat_threshold)
        return res

    def get_mutated_genes(self):
        """Return dict of gene mutation frequencies based on TCGA studies."""
        if self.mutation_cache:
            logger.info('Loading mutations from %s' % self.mutation_cache)
            with open(self.mutation_cache, 'r') as fh:
                self.mutations = json.load(fh)
        else:
            logger.info('Getting mutations from cBio web service')
            mutations = {}
            for tcga_study_name in tcga_studies[self.tcga_study_prefix]:
                for idx, hgnc_name_batch in \
                                enumerate(batch_iter(hgnc_ids.keys(), 200)):
                    logger.info('Fetching mutations for %s and gene batch %s' %
                                (tcga_study_name, idx))
                    patient_mutations = \
                        cbio_client.get_profile_data(tcga_study_name,
                                                     hgnc_name_batch,
                                                     'mutation')
                    # e.g. 'ICGC_0002_TD': {'BRAF': None, 'KRAS': 'G12D'}
                    for patient, gene_mut_dict in patient_mutations.items():
                        # 'BRAF': None
                        for gene, mutated in gene_mut_dict.items():
                            if mutated is not None:
                                try:
                                    mutations[gene] += 1
                                except KeyError:
                                    mutations[gene] = 1
            self.mutations = mutations

        # Normalize mutations by length
        self.norm_mutations = {}
        for gene_name, num_muts in self.mutations.items():
            self.norm_mutations[gene_name] = \
                self.normalize_mutation_count(gene_name, num_muts)

        return self.mutations, self.norm_mutations

    @staticmethod
    def normalize_mutation_count(gene_name, num_muts):
        hgnc_id = get_hgnc_id(gene_name)
        up_id = get_uniprot_id(hgnc_id)
        if not up_id:
            logger.warning("Could not get Uniprot ID for HGNC symbol %s "
                           "with HGNC ID %s" % (gene_name, hgnc_id))
            length = 500 # a guess at a default
        else:
            length = uniprot_client.get_length(up_id)
            if not length:
                logger.warning("Could not get length for Uniprot "
                               "ID %s" % up_id)
                length = 500 # a guess at a default
        norm_mutations = num_muts / float(length)
        return norm_mutations

    def load_sif_prior(self, fname, e50=20):
        """Return a Graph based on a SIF file describing a prior.

        Parameters
        ----------
        fname : str
            Path to the SIF file.
        e50 : int
            Parameter for converting evidence counts into weights over the
            interval [0, 1) according to hyperbolic function
            `weight = (count / (count + e50))`.
        """
        # Format
        # agA_ns,agA_id,agA_name,agB_ns,agB_id,agB_name,stmt_type,
        #   evidence_count
        # FPLX,9_1_1,9_1_1,HGNC,3511,EXO1,Activation,7
        G = nx.Graph()
        logger.info('Loading SIF prior from %s' % fname)
        with open(fname, 'r') as fh:
            csv_reader = csv.reader(fh, delimiter=',')
            header = next(csv_reader)
            for row in csv_reader:
                agA_ns, agA_id, agA_name, agB_ns, agB_id, agB_name, \
                    stmt_type, evidence_count = row
                A_key = '%s:%s' % (agA_ns, agA_id)
                B_key = '%s:%s' % (agB_ns, agB_id)
                weight = (float(evidence_count) /
                          (e50 + float(evidence_count)))
                G.add_edge(A_key, B_key, weight=weight)
        self.prior_graph = G
        logger.info('Finished loading SIF prior')
        return G

    def get_relevant_nodes(self, pct_heat_threshold):
        """Return a list of the relevant nodes in the prior.

        Heat diffusion is applied to the prior network based on initial
        heat on nodes that are mutated according to patient statistics.
        """
        logger.info('Setting heat for relevant nodes in prior network')
        heats = np.zeros(len(self.prior_graph))
        mut_nodes = {}
        for gene_name, muts in self.norm_mutations.items():
            if muts:
                hgnc_id = get_hgnc_id(gene_name)
                node_key = 'HGNC:%s' % hgnc_id
                mut_nodes[node_key] = muts

        for idx, node in enumerate(self.prior_graph.nodes()):
            if node in mut_nodes:
                heats[idx] = mut_nodes[node]

        gamma = -0.1
        logger.info('Calculating Laplacian matrix')
        lp_mx = nx.normalized_laplacian_matrix(self.prior_graph,
                                               weight='weight')
        logger.info('Diffusing heat')
        Df = expm_multiply(gamma * lp_mx, heats)
        heat_thresh = np.percentile(Df, pct_heat_threshold)
        logger.info('Filtering to relevant nodes with heat threshold %.2f '
                    '(%s percentile)' % (heat_thresh, pct_heat_threshold))
        # Zip the nodes with their heats and sort
        node_heats = sorted(list(zip(self.prior_graph.nodes(), Df)),
                            key=lambda x: x[1], reverse=True)
        relevant_nodes = [n for n, heat in node_heats if heat >= heat_thresh]
        return relevant_nodes

    @staticmethod
    def search_terms_from_nodes(node_list):
        """Build a list of Pubmed search terms from the nodes returned by
        make_prior."""
        terms = []
        for node in node_list:
            if node.startswith('HGNC:'):
                hgnc_id = node.split(':')[1]
                hgnc_name = get_hgnc_name(hgnc_id)
                if hgnc_name is None:
                    logger.log(f'{node} is not a valid HGNC ID')
                else:
                    term = SearchTerm(type='gene', name=hgnc_name,
                                      search_term=f'"{hgnc_name}"',
                                      db_refs={'HGNC': hgnc_id})
                    terms.append(term)
            elif node.startswith('MESH:'):
                mesh_id = node.split(':')[1]
                # TODO: get actual process name here
                term = SearchTerm(type='bioprocess', name=mesh_id,
                                  search_term=f'{mesh_id}[MeSH Terms]',
                                  db_refs={'MESH': mesh_id})
                terms.append(term)
            # TODO: handle GO here
            else:
                logger.warning(f'Could not create search term from {node}')
        return sorted(terms, key=lambda x: x.name)

    @staticmethod
    def find_drugs_for_genes(node_list):
        """Return list of drugs targeting gene nodes."""
        tas_statements = tas.process_csv().statements
        already_added = set()
        drug_terms = []
        for node in node_list:
            if node.startswith('HGNC:'):
                hgnc_id = node.split(':')[1]
                drugs = get_drugs_for_gene(tas_statements, hgnc_id)
                for drug in drugs:
                    if drug.name not in already_added:
                        drug_terms.append(drug)
                        already_added.add(drug.name)
        return sorted(drug_terms, key=lambda x: x.name)


def _load_tcga_studies():
    """Return a list of TCGA studies by prefix.

    Note that the resource file read here ensures that studies are
    non-redundant, which wouldn't be guaranteed by the web service.
    """
    here = os.path.dirname(os.path.abspath(__file__))
    resources = os.path.join(here, os.pardir, 'resources')
    studies_file = os.path.join(resources, 'cancer_studies.json')
    with open(studies_file, 'r') as fh:
        studies = json.load(fh)
    return studies


tcga_studies = _load_tcga_studies()



import time
import random

from nose.plugins.attrib import attr

from emmaa.db import Query, Result, EmmaaDatabaseManager
from emmaa.queries import Query as QueryObject, PathProperty


test_query_jsons = [{'type': 'path_property', 'path': {'type': 'Activation',
                     'subj': {'type': 'Agent', 'name': 'BRAF'},
                     'obj': {'type': 'Agent', 'name': 'ERK'}}},
                    {'type': 'path_property', 'path':
                    {'type': 'Phosphorylation',
                     'enz': {'type': 'Agent', 'name': 'ERK'},
                     'sub': {'type': 'Agent', 'name': 'MEK'}}}]

test_queries = [QueryObject._from_json(qj) for qj in test_query_jsons]


def _get_test_db():
    db = EmmaaDatabaseManager('postgresql://postgres:@localhost/emmaadb_test')
    db.drop_tables(force=True)
    db.create_tables()
    return db


@attr('nonpublic')
def test_instantiation():
    db = _get_test_db()
    assert db
    return


@attr('nonpublic')
def test_put_queries():
    db = _get_test_db()
    db.put_queries('joshua', 1, test_queries[0], ['aml', 'luad'])
    with db.get_session() as sess:
        queries = sess.query(Query).all()
    assert len(queries) == 2, len(queries)


@attr('nonpublic')
def test_get_queries():
    db = _get_test_db()
    for query in test_queries:
        db.put_queries('joshua', 1, query, ['aml', 'luad'])
    queries = db.get_queries('aml')
    assert len(queries) == 2, len(queries)
    assert all(isinstance(query, PathProperty) for query in queries)


def _get_random_result():
    return random.choice(
        [{'12': [['This is fine.', '']]}, {'34': [['This is not ok.', '']]}])


@attr('nonpublic')
def test_put_results():
    db = _get_test_db()
    db.put_queries('joshua', 1, test_queries[0], ['aml', 'luad'])
    queries = db.get_queries('aml')
    results = [(query, '', _get_random_result())
               for query in queries]
    db.put_results('aml', results)
    with db.get_session() as sess:
        db_results = sess.query(Result).all()
    assert len(db_results) == len(results)


@attr('nonpublic')
def test_get_results():
    db = _get_test_db()
    models = ['aml', 'luad']

    # Fill up the database.
    for query in test_queries:
        db.put_queries('joshua', 1, query, models)
    for model in models:
        db.put_results(model, [(query, 'pysb', _get_random_result())
                               for query in test_queries])

    # Try to get the results.
    results = db.get_results('joshua')
    assert len(results) == len(test_queries)*len(models), len(results)
    assert all(isinstance(result, tuple) for result in results)
    assert all(result[0] in models for result in results)
    assert any(results[0][1].matches(q) for q in test_queries)
    assert all(isinstance(result[2], str) for result in results)
    assert all(isinstance(result[3], dict) for result in results)


@attr('nonpublic')
def test_get_latest_results():
    db = _get_test_db()
    models = ['aml', 'luad']

    # Fill up the database.
    for query in test_queries:
        db.put_queries('joshua', 1, query, models)
    for model in models:
        db.put_results(model, [(query, 'pysb', _get_random_result())
                               for query in test_queries])

    # Add the same statements over again
    time.sleep(10)
    for model in models:
        db.put_results(model, [(query, 'pysb', _get_random_result())
                               for query in test_queries])

    # Try to get the results. Make sure we only get the one set.
    results = db.get_results('joshua')
    assert len(results) == len(test_queries)*len(models), len(results)
    assert all(isinstance(result, tuple) for result in results)
    assert all(result[0] in models for result in results)
    assert any(results[0][1].matches(q) for q in test_queries)
    assert all(isinstance(result[2], str) for result in results)
    assert all(isinstance(result[3], dict) for result in results)



import re
import logging
import requests
from functools import lru_cache

from indra.sources import tas
from indra.databases.uniprot_client import get_gene_name
from indra.databases.hgnc_client import get_hgnc_id, get_uniprot_id

from emmaa.priors import get_drugs_for_gene, SearchTerm

logger = logging.getLogger('reactome_prior')


def make_prior_from_genes(gene_list):
    """Return reactome prior based on a list of genes

    Parameters
    ----------
    gene_list : list of str
        List of HGNC symbols for genes

    Returns
    -------
    res : list of :py:class:`emmaa.priors.SearchTerm`
        List of search terms corresponding to all genes found in any reactome
        pathway containing one of the genes in the input gene list
    """
    all_reactome_ids = set([])
    for gene_name in gene_list:
        hgnc_id = get_hgnc_id(gene_name)
        uniprot_id = get_uniprot_id(hgnc_id)
        if not uniprot_id:
            logger.warning('Could not get Uniprot ID for HGNC symbol'
                           f' {gene_name}')
            continue
        reactome_ids = rx_id_from_up_id(uniprot_id)
        if not reactome_ids:
            logger.warning('Could not get Reactome ID for Uniprot ID'
                           f' {uniprot_id} with corresonding HGNC symbol'
                           f' {gene_name}')
            continue
        all_reactome_ids.update(reactome_ids)

    all_pathways = set([])
    for reactome_id in all_reactome_ids:
        if not re.match('^R-HSA-[0-9]', reactome_id):
            # skip non-human genes
            continue
        additional_pathways = get_pathways_containing_gene(reactome_id)
        if additional_pathways is not None:
            all_pathways.update(additional_pathways)

    all_genes = set([])
    for pathway in all_pathways:
        additional_genes = get_genes_contained_in_pathway(pathway)
        if additional_genes is not None:
            all_genes.update(additional_genes)

    gene_terms = []
    for uniprot_id in all_genes:
        hgnc_name = get_gene_name(uniprot_id)
        if hgnc_name is None:
            logger.warning('Could not get HGNC name for UniProt ID'
                           f' {uniprot_id}')
            continue
        hgnc_id = get_hgnc_id(hgnc_name)
        if not hgnc_id:
            logger.warning('Could not find HGNC ID for HGNC symbol'
                           f' {hgnc_name} with corresonding Uniprot ID'
                           f' {uniprot_id}')
            continue
        term = SearchTerm(type='gene', name=hgnc_name,
                          search_term=f'"{hgnc_name}"',
                          db_refs={'HGNC': hgnc_id,
                                   'UP': uniprot_id})
        gene_terms.append(term)
    return sorted(gene_terms, key=lambda x: x.name)


def find_drugs_for_genes(search_terms):
    """Return list of drugs targeting at least one gene from a list of genes

    Parameters
    ----------
    search_terms : list of :py:class:`emmaa.priors.SearchTerm`
        List of search terms for genes

    Returns
    -------
    drug_terms : list of :py:class:`emmaa.priors.SearchTerm`
        List of search terms of drugs targeting at least one of the input genes
    """
    tas_statements = tas.process_csv().statements
    drug_terms = []
    already_added = set()
    for search_term in search_terms:
        if search_term.type == 'gene':
            hgnc_id = search_term.db_refs['HGNC']
            drugs = get_drugs_for_gene(tas_statements, hgnc_id)
            for drug in drugs:
                if drug.name not in already_added:
                    drug_terms.append(drug)
                    already_added.add(drug.name)
    return sorted(drug_terms, key=lambda x: x.name)


@lru_cache(10000)
def rx_id_from_up_id(up_id):
    """Return the Reactome Stable IDs for a given Uniprot ID."""
    react_search_url = 'http://www.reactome.org/ContentService/search/query'
    params = {'query': up_id, 'cluster': 'true', 'species': 'Homo sapiens'}
    headers = {'Accept': 'application/json'}
    res = requests.get(react_search_url, headers=headers, params=params)
    if not res.status_code == 200:
        return None
    json = res.json()
    results = json.get('results')
    if not results:
        logger.warning(f'No results for {up_id}')
        return None
    stable_ids = []
    for result in results:
        entries = result.get('entries')
        for entry in entries:
            stable_id = entry.get('stId')
            if not stable_id:
                continue
            stable_ids.append(stable_id)
    return stable_ids


@lru_cache(100000)
def up_id_from_rx_id(reactome_id):
    """Get the Uniprot ID (referenceEntity) for a given Reactome Stable ID."""
    react_url = 'http://www.reactome.org/ContentService/data/query/' \
                + reactome_id + '/referenceEntity'
    res = requests.get(react_url)
    if not res.status_code == 200:
        return None
    _, entry, entry_type = res.text.split('\t')
    if entry_type != 'ReferenceGeneProduct':
        return None
    id_entry = entry.split(' ')[0]
    db_ns, db_id = id_entry.split(':')
    if db_ns != 'UniProt':
        return None
    return db_id


@lru_cache(1000)
def get_pathways_containing_gene(reactome_id):
    """"Get all ids for reactom pathways containing some form of an entity

    Parameters
    ----------
    reactome_id : str
        Reactome id for a gene

    Returns
    -------
    pathway_ids : list of str
        List of reactome ids for pathways containing the input gene
    """
    react_url = ('http://www.reactome.org/ContentService/data/pathways/low'
                 f'/entity/{reactome_id}/allForms')
    params = {'species': 'Homo sapiens'}
    headers = {'Accept': 'application/json'}
    res = requests.get(react_url, headers=headers, params=params)
    if not res.status_code == 200:
        logger.warning(f'Request failed for reactome_id {reactome_id}')
        return None
    results = res.json()
    if not results:
        logger.info(f'No results for {reactome_id}')
        return None
    pathway_ids = [pathway['stIdVersion'] for pathway in results]
    return pathway_ids


@lru_cache(1000)
def get_genes_contained_in_pathway(reactome_id):
    """Get all genes contained in a given pathway

    Parameters
    ----------
    reactome_id : str
        Reactome id for a pathway

    Returns
    -------
    genes : list of str
        List of uniprot ids for all unique genes contained in input pathway
    """
    react_url = ('http://www.reactome.org/ContentService/data'
                 f'/participants/{reactome_id}')
    params = {'species': 'Homo species'}
    headers = {'Accept': 'application/json'}
    res = requests.get(react_url, headers=headers, params=params)
    results = res.json()
    if not res.status_code == 200:
        return None
    if not results:
        logger.info(f'No results for {reactome_id}')
    genes = [entity['identifier'] for result in results
             for entity in result['refEntities']
             if entity.get('schemaClass') == 'ReferenceGeneProduct']
    return list(set(genes))



import boto3
import pickle
import unittest
from indra.tools.reading.submit_reading_pipeline import wait_for_complete

from emmaa.aws_lambda_functions.analyze_changes_on_s3 import lambda_handler, QUEUE
from emmaa.util import make_date_str, get_s3_client

RUN_STATI = ['SUBMITTED', 'PENDING', 'RUNNABLE', 'RUNNING']
DONE_STATI = ['SUCCEEDED', 'FAILED']


def __get_jobs(batch):
    job_ids = {}
    for status in RUN_STATI + DONE_STATI:
        resp = batch.list_jobs(jobQueue=QUEUE, jobStatus=status)
        if 'jobSummaryList' in resp.keys():
            job_ids[status] = [s['jobId'] for s in resp['jobSummaryList']]
    return job_ids


@unittest.skip('Local test without starting up batch job not yet implemented')
def test_handler():
    """Test the lambda handler locally."""
    dts = make_date_str()
    key = f'models/test/test_model_{dts}.pkl'
    event = {'Records': [{'s3': {'object': {'key': key}}}]}
    context = None
    res = lambda_handler(event, context)
    print(res)
    assert res['statusCode'] == 200, res
    assert res['result'] == 'SUCCESS', res
    assert res['job_id'], res
    job_id = res['job_id']

    results = {}
    wait_for_complete(QUEUE, job_list=[{'jobId': job_id}],
                      result_record=results)
    print(results)
    assert job_id in [job_def['jobId'] for job_def in results['succeeded']], \
        results['failed']

    s3 = get_s3_client()
    s3_res = s3.list_objects(Bucket='emmaa', Prefix='results/test/' + dts[:10])
    print(s3_res.keys())
    assert s3_res, s3_res


@unittest.skip('Unfinished test. See comments in code')
def test_s3_response():
    """Change a file on s3 and check for the correct response."""
    # This will be a white-box test. We will check progress at various stages.
    s3 = get_s3_client()
    batch = boto3.client('batch')

    # Define some fairly random parameters.
    key = f'models/test/model_{make_date_str()}.pkl'
    data = {'test_message': 'Hello world!'}

    # This should trigger the lambda to start a batch job.
    s3.put_object(Bucket='emmaa', Key=key, Body=pickle.dumps(data))

    # TODO
    #  1. verify that lambda has started a batch job
    #  2. kill batch job
    #  3. delete uploaded pickle



from indra.sources import tas
from indra.statements import Agent
from indra.databases import hgnc_client
from . import SearchTerm, get_drugs_for_gene
from . prior_stmts import get_stmts_for_gene_list
import datetime
from emmaa.statements import EmmaaStatement
from emmaa.model import EmmaaModel, save_config_to_s3


class GeneListPrior(object):
    """Class to manage the construction of a model from a list of genes.

    Parameters
    ----------
    gene_list : list[str]
        A list of HGNC gene symbols
    name : str
        The name of the model (all lower case, no spaces or special characters)
    human_readable_name : str
        The human readable name (display name) of the model
    """
    def __init__(self, gene_list, name, human_readable_name):
        self.name = name
        self.gene_list = gene_list
        self.human_readable_name = human_readable_name
        self.stmts = []
        self.search_terms = []

    def make_search_terms(self):
        """Generate search terms from the gene list."""
        tas_stmts = tas.process_csv().statements
        already_added = set()
        terms = []
        for gene in self.gene_list:
            # Gene search term
            agent = agent_from_gene_name(gene)
            term = SearchTerm(type='gene', name=agent.name,
                              search_term=f'"{agent.name}"',
                              db_refs={'HGNC': agent.db_refs['HGNC'],
                                       'UP': agent.db_refs['UP']})
            terms.append(term)

            # Drug search term
            drug_terms = get_drugs_for_gene(tas_stmts,
                                            agent.db_refs['HGNC'])
            for drug_term in drug_terms:
                if drug_term.name not in already_added:
                    terms.append(drug_term)
                    already_added.add(drug_term.name)
        self.search_terms = terms
        return terms

    def make_gene_statements(self):
        """Generate Statements from the gene list."""
        drug_names = [st.name for st in self.search_terms if
                      st.type == 'drug']
        indra_stmts = get_stmts_for_gene_list(self.gene_list, drug_names)
        estmts = [EmmaaStatement(stmt, datetime.datetime.now(), [])
                  for stmt in indra_stmts]
        self.stmts = estmts

    def make_config(self):
        """Generate a configuration based on attributes."""
        if not self.search_terms:
            self.make_search_terms()
        if not self.stmts:
            self.make_gene_statements()
        config = dict()
        config['name'] = self.name
        config['human_readable_name'] = self.human_readable_name
        config['search_terms'] = [st.to_json() for st in self.search_terms]
        config['assembly'] = {
            'belief_cutoff': 0.8,
            'filter_ungrounded': True
        }
        return config

    def make_model(self):
        """Make an EmmaaModel and upload it along with the config to S3."""
        config = self.make_config()
        em = EmmaaModel(self.name, config)
        em.stmts = self.stmts
        ndex_uuid = em.upload_to_ndex()
        config['ndex'] = {'network': ndex_uuid}
        save_config_to_s3(self.name, config)
        em.save_to_s3()


def agent_from_gene_name(gene_name):
    """Return an Agent based on a gene name."""
    hgnc_id = hgnc_client.get_hgnc_id(gene_name)
    up_id = hgnc_client.get_uniprot_id(hgnc_id)
    agent = Agent(gene_name, db_refs={'HGNC': hgnc_id,
                                      'UP': up_id})
    return agent



import os
import logging
import datetime
from indra.sources import eidos
from indra.literature import elsevier_client
from emmaa.statements import EmmaaStatement


logger = logging.getLogger(__name__)


def read_elsevier_eidos_search_terms(piis_to_terms):
    """Return extracted EmmaaStatements given a dict of PIIS to SearchTerms.

    Parameters
    ----------
    piis_to_terms : dict
        A dict representing a set of PIIs pointing to search terms that
        produced them.

    Returns
    -------
    list[:py:class:`emmaa.model.EmmaaStatement`]
        A list of EmmaaStatements extracted from the given PMIDs.
    """
    piis = list(piis_to_terms.keys())
    date = datetime.datetime.utcnow()
    texts = read_piis(piis)
    pii_stmts = process_texts(texts)
    estmts = []
    for pii, stmts in pii_stmts.items():
        for stmt in stmts:
            es = EmmaaStatement(stmt, date, piis_to_terms[pii])
            estmts.append(es)
    return estmts


def read_piis(piis):
    """Return texts extracted from articles with given PIIs.

    Parameters
    ----------
    piis : list[str]
        A list of PIIs to extract texts from.

    Returns
    -------
    texts : dict
        A dictionary representing PIIs as keys and extracted texts as values.
    """
    texts = {}
    for pii in piis:
        try:
            xml = elsevier_client.download_article(pii, id_type='pii')
            # If we got an empty xml or bad response
            if not xml:
                logger.info('Could not get article content for %s' % pii)
                continue
        # Handle Connection and other errors
        except Exception as e:
            logger.info('Could not get article content for %s because of %s'
                        % (pii, e))
            continue
        try:
            txt = elsevier_client.extract_text(xml)
            # If we could find relevant xml parts
            if not txt:
                logger.info('Could not extract article text for %s' % pii)
                continue
        # Handle Connection and other errors
        except Exception as e:
            logger.info('Could not extract article text for %s because of %s'
                        % (pii, e))
        texts[pii] = txt
    logger.info('Got text back for %d articles.' % len(texts))
    return texts


def process_texts(texts):
    """Process article texts with Eidos and extract INDRA Statements.

    Parameters
    ----------
    texts : dict
        A dictionary mapping PIIs to texts to process.

    Returns
    -------
    pii_stmts : dict
        A dictionary mapping PIIs as keys and extracted INDRA statements.
    """
    eidos_url = os.environ.get('EIDOS_URL')
    logger.info('Reading with Eidos URL: %s' % eidos_url)
    pii_stmts = {}
    for pii, txt in texts.items():
        logger.info('Reading the article with %s pii.' % pii)
        try:
            ep = eidos.process_text(txt, webservice=eidos_url)
            if ep:
                pii_stmts[pii] = ep.statements
        # Handle Connection and other errors
        except Exception as e:
            logger.info('Could not read the text because of %s' % str(e))
            continue
    return pii_stmts



"""The AWS Lambda update definition.

This file contains the function that will be run when Lambda is triggered. It
must be placed on s3, which can either be done manually (not recommended) or
by running:

$ python update_lambda.py analyze_changes_on_s3.py emmaa-analysis

in this directory.
"""

import boto3
from datetime import datetime

JOB_DEF = 'emmaa_jobdef'
QUEUE = 'emmaa-models-update-test'
PROJECT = 'aske'
PURPOSE = 'update-emmaa-results'
BRANCH = 'origin/master'


def lambda_handler(event, context):
    """Initial batch jobs for any changed models or tests.

    This function is designed to be placed on lambda, taking the event and
    context arguments that are passed, and extracting the names of the
    uploaded (which includes changed) model or test definitions on s3.
    Lambda is configured to be triggered by any such changes, and will
    automatically run this script.

    See the top of the page for the Lambda update procedure.

    Note that this function must always have the same parameters, even if any
    or all of them are unused, because we do not have control over what Lambda
    sends as parameters. For more information on the context parameter, see
    here:

      https://docs.aws.amazon.com/lambda/latest/dg/python-context-object.html

    And for more informationn about AWS Lambda handlers, see here:

      https://docs.aws.amazon.com/lambda/latest/dg/python-programming-model-handler-types.html


    Parameters
    ----------
    event : dict
        A dictionary containing metadata regarding the triggering event. In
        this case, we are expecting 'Records', each of which contains a record
        of a file that was added (or changed) on s3.
    context : object
        This is an object containing potentially useful context provided by
        Lambda. See the documentation cited above for details.

    Returns
    -------
    ret : dict
        A dict containing 'statusCode', with a valid HTTP status code, and any
        other data to be returned to Lambda.
    """
    batch = boto3.client('batch')
    records = event['Records']
    for rec in records:
        try:
            model_key = rec['s3']['object']['key']
        except KeyError:
            pass
        model_name = model_key.split('/')[1]
        core_command = 'bash scripts/git_and_run.sh'
        if BRANCH is not None:
            core_command += f' --branch {BRANCH}'
        core_command += (' python scripts/run_model_tests_from_s3.py'
                         f' --model {model_name}')
        print(core_command)
        cont_overrides = {
            'command': ['python', '-m', 'indra.util.aws', 'run_in_batch',
                        '--project', PROJECT, '--purpose', PURPOSE,
                        core_command]
            }
        now_str = datetime.utcnow().strftime('%Y%m%d_%H%M%S')
        ret = batch.submit_job(jobName=f'{model_name}_{now_str}',
                               jobQueue=QUEUE, jobDefinition=JOB_DEF,
                               containerOverrides=cont_overrides)
        job_id = ret['jobId']

    return {'statusCode': 200, 'result': 'SUCCESS', 'job_id': job_id}



import re

from emmaa.util import RE_DATEFORMAT
from emmaa_service.api import get_model_stats, model_last_updated

MODEL = 'aml'


def test_last_updated():
    key_str = model_last_updated(MODEL)
    assert key_str
    assert re.search(RE_DATEFORMAT, key_str).group()


def test_get_model_statistics():
    model_json = get_model_stats(MODEL)
    assert isinstance(model_json, dict)



import logging

logging_format = '%(levelname)s: [%(asctime)s] %(name)s - %(message)s'
logging.basicConfig(format=logging_format, level=logging.INFO,
                    datefmt='%Y-%m-%d %H:%M:%S')

logger = logging.getLogger('emmaa')



import re
import unittest
from emmaa.priors import SearchTerm
from emmaa.priors.reactome_prior import rx_id_from_up_id
from emmaa.priors.reactome_prior import find_drugs_for_genes
from emmaa.priors.reactome_prior import make_prior_from_genes
from emmaa.priors.reactome_prior import get_pathways_containing_gene
from emmaa.priors.reactome_prior import get_genes_contained_in_pathway


def test_rx_id_from_up_id():
    """Check that Uniprot ids are being successfully mapped to reactome ids
    """
    test_cases = [('P01116', 'R-HSA-9653079'),   # KRAS
                  ('P04637', 'R-HSA-69507'),   # TP53
                  ('Q13485', 'R-HSA-2187323')]  # SMAD4
    for up_id, rx_id in test_cases:
        all_rx_ids = rx_id_from_up_id(up_id)
        assert rx_id in all_rx_ids


def test_get_pathways_containing_genes():
    # get pathways containing KRAS
    KRAS_pathways = get_pathways_containing_gene('R-HSA-62719')
    pattern = r'R-HSA-[0-9]{7}\.1$'
    # Check if function returns a list of valid pathway ids
    assert all([re.match(pattern, pathway_id) for pathway_id in KRAS_pathways])
    # Check if function returns a reasonable number of pathways
    assert len(KRAS_pathways) > 3
    # Signaling Downstream of RAS mutants
    assert 'R-HSA-9649948.1' in KRAS_pathways
    

def test_get_genes_contained_in_pathway():
    # Get genes in Signaling by RAS mutants pathway
    RAS_mutants_genes = get_genes_contained_in_pathway('R-HSA-6802949.1')
    pattern = re.compile(r'(O|P|Q)[A-Z0-9]{5}$')
    # Check that function produces list of valid human uniprot ids
    assert all([re.match(pattern, up_id) for up_id in RAS_mutants_genes])
    # Check that function returns a reasonable number of genes
    assert len(RAS_mutants_genes) > 30
    # KRAS
    assert 'P01116' in RAS_mutants_genes


def test_make_prior_from_genes():
    # KRAS prior
    prior1 = make_prior_from_genes(['KRAS'])
    # BRCA prior
    prior2 = make_prior_from_genes(['TP53', 'PIK3CA', 'GATA3', 'CBFB', 'CDH1'])
    # make sure there are results
    assert prior1
    # make sure the prior is a list of SearchTerms
    assert all(isinstance(term, SearchTerm) for term in prior1)
    # if we get fewer than 40 genes for KRAS it's likely something is wrong
    assert len(prior1) > 40
    # test that the prior contains some of the usual suspects
    gene_names = set(term.name for term in prior1)
    assert set(['KRAS', 'RAF1', 'MAPK1', 'BRAF']) <= gene_names

    assert prior2
    assert all(isinstance(term, SearchTerm) for term in prior2)
    assert len(prior2) > 1000
    gene_names = set(term.name for term in prior2)
    assert set(['COL1A1', 'ESR1', 'EGFR', 'HRAS']) <= gene_names
    # some genes expressed only in the brain
    assert not (gene_names & set(['BARHL1', 'NEUROD2']))


def test_find_drugs_for_genes():
    # SearchTerm for SRC
    SRC = SearchTerm(type='gene', name='SRC', search_term='"SRC"',
                     db_refs={'HGNC': '11283'})
    # drugs targeting KRAS
    drug_terms = find_drugs_for_genes([SRC])

    # make sure there are results
    assert drug_terms

    # make sure the result is a list of search terms
    assert all(isinstance(term, SearchTerm) for term in drug_terms)

    # something is wrong if there are fewer than 10 drugs
    assert len(drug_terms) > 10

    # test that some example drugs are included
    drug_names = set(term.name for term in drug_terms)
    example_drugs = set(['Dasatinib', 'Tozasertib', 'Ponatinib'])
    assert example_drugs <= drug_names
